"""_stratigraphic_unit_feature.py: RESQML StratigraphicUnitFeature class."""

# NB: in this module, the term 'unit' refers to a geological stratigraphic unit, i.e. a layer of rock, not a unit of measure

import logging

log = logging.getLogger(__name__)

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.organize as rqo
from resqpy.olio.base import BaseResqpy


class StratigraphicUnitFeature(BaseResqpy):
    """Class for RESQML Stratigraphic Unit Feature objects.

    RESQML documentation:

       A stratigraphic unit that can have a well-known (e.g., "Jurassic") chronostratigraphic top and
       chronostratigraphic bottom. These chronostratigraphic units have no associated interpretations or representations.

       BUSINESS RULE: The name must reference a well-known chronostratigraphic unit (such as "Jurassic"),
       for example, from the International Commission on Stratigraphy (http://www.stratigraphy.org).
    """

    resqml_type = 'StratigraphicUnitFeature'

    def __init__(self,
                 parent_model,
                 uuid = None,
                 top_unit_uuid = None,
                 bottom_unit_uuid = None,
                 title = None,
                 originator = None,
                 extra_metadata = None):
        """Initialises a stratigraphic unit feature object.

        arguments:
           parent_model (model.Model): the model with which the new feature will be associated
           uuid (uuid.UUID, optional): the uuid of an existing RESQML stratigraphic unit feature from which
              this object will be initialised
           top_unit_uuid (uuid.UUID, optional): the uuid of a geologic or stratigraphic unit feature which is
              at the top of the new unit; ignored if uuid is not None
           bottom_unit_uuid (uuid.UUID, optional): the uuid of a geologic or stratigraphic unit feature which is
              at the bottom of the new unit; ignored if uuid is not None
           title (str, optional): the citation title (feature name) of the new feature; ignored if uuid is not None
           originator (str, optional): the name of the person creating the new feature; ignored if uuid is not None
           extra_metadata (dict, optional): extra metadata items for the new feature

        returns:
           a new stratigraphic unit feature resqpy object
        """

        # todo: clarify with Energistics whether the 2 references are to other StratigraphicUnitFeatures or what?
        self.top_unit_uuid = top_unit_uuid
        self.bottom_unit_uuid = bottom_unit_uuid

        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

    def is_equivalent(self, other, check_extra_metadata = True):
        """Returns True if this feature is essentially the same as the other; otherwise False.

        arguments:
           other (StratigraphicUnitFeature): the other feature to compare this one against
           check_extra_metadata (bool, default True): if True, then extra metadata items must match for the two
              features to be deemed equivalent; if False, extra metadata is ignored in the comparison

        returns:
           bool: True if this feature is essentially the same as the other feature; False otherwise
        """

        if not isinstance(other, StratigraphicUnitFeature):
            return False
        if self is other or bu.matching_uuids(self.uuid, other.uuid):
            return True
        return (self.title == other.title and
                ((not check_extra_metadata) or rqo.equivalent_extra_metadata(self, other)))

    def _load_from_xml(self):
        """Loads class specific attributes from xml for an existing RESQML object; called from BaseResqpy."""
        root_node = self.root
        assert root_node is not None
        bottom_ref_uuid = rqet.find_nested_tags_text(root_node, ['ChronostratigraphicBottom', 'UUID'])
        top_ref_uuid = rqet.find_nested_tags_text(root_node, ['ChronostratigraphicTop', 'UUID'])
        # todo: find out if these are meant to be references to other stratigraphic unit features or geologic unit features
        # and if so, instantiate those objects?
        # for now, simply note the uuids
        if bottom_ref_uuid is not None:
            self.bottom_unit_uuid = bu.uuid_from_string(bottom_ref_uuid)
        if top_ref_uuid is not None:
            self.top_unit_uuid = bu.uuid_from_string(top_ref_uuid)

    def create_xml(self, add_as_part = True, originator = None, reuse = True, add_relationships = True):
        """Creates xml for this stratigraphic unit feature.

        arguments:
           add_as_part (bool, default True): if True, the feature is added to the parent model as a high level part
           originator (str, optional): if present, is used as the originator field of the citation block
           reuse (bool, default True): if True, the parent model is inspected for any equivalent feature and, if found,
              the uuid of this feature is set to that of the equivalent part
           add_relationships (bool, default True): if True and add_as_part is True, relationships are created with
              the referenced top and bottom units, if present

        returns:
           lxml.etree._Element: the root node of the newly created xml tree for the feature
        """

        if reuse and self.try_reuse():
            return self.root  # check for reusable (equivalent) object
        # create node with citation block
        suf = super().create_xml(add_as_part = False, originator = originator)

        if self.bottom_unit_uuid is not None:
            self.model.create_ref_node('ChronostratigraphicBottom',
                                       self.model.title(uuid = self.bottom_unit_uuid),
                                       self.bottom_unit_uuid,
                                       content_type = self.model.type_of_uuid(self.bottom_unit_uuid),
                                       root = suf)

        if self.top_unit_uuid is not None:
            self.model.create_ref_node('ChronostratigraphicTop',
                                       self.model.title(uuid = self.top_unit_uuid),
                                       self.top_unit_uuid,
                                       content_type = self.model.type_of_uuid(self.top_unit_uuid),
                                       root = suf)

        if add_as_part:
            self.model.add_part('obj_StratigraphicUnitFeature', self.uuid, suf)
            if add_relationships:
                if self.bottom_unit_uuid is not None:
                    self.model.create_reciprocal_relationship(suf, 'destinationObject',
                                                              self.model.root(uuid = self.bottom_unit_uuid),
                                                              'sourceObject')
                if self.top_unit_uuid is not None and not bu.matching_uuids(self.bottom_unit_uuid, self.top_unit_uuid):
                    self.model.create_reciprocal_relationship(suf, 'destinationObject',
                                                              self.model.root(uuid = self.top_unit_uuid),
                                                              'sourceObject')

        return suf



"""A submodule containing grid transmissibility functions"""

# note: only IJK Grid format supported at present
# see also rq_import.py

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.olio.transmission as rqtr

always_write_pillar_geometry_is_defined_array = False
always_write_cell_geometry_is_defined_array = False


def transmissibility(grid, tolerance = 1.0e-6, use_tr_properties = True, realization = None, modifier_mode = None):
    """Returns transmissibilities for standard (IJK neighbouring) connections within this grid.

    arguments:
       tolerance (float, default 1.0e-6): the minimum half cell transmissibility below which zero inter-cell
          transmissibility will be set; units are as for returned values (see notes)
       use_tr_properties (boolean, default True): if True, the grid's property collection is inspected for
          possible transmissibility arrays and if found, they are used instead of calculation; note that
          when this argument is False, the property collection is still used for the feed arrays to the
          calculation
       realization (int, optional) if present, only properties with this realization number will be used;
          applies to pre-computed transmissibility properties or permeability and net to gross ratio
          properties when computing
       modifier_mode (string, optional): if None, no transmissibility modifiers are applied; other
          options are: 'faces multiplier', for which transmissibility properties with indexable
          element of 'faces' will be used (directional or composite); 'faces per cell multiplier',
          in which case a transmissibility property with 'faces per cell' as the indexable element
          will be used to modify the half cell transmissibilities prior to combination;
          or 'absolute' in which case directional properties of local property kind 'fault transmissibility'
          (or 'mat transmissibility') and indexable element of 'faces' will be used as a third transmissibility
          term along with the two half cell transmissibilities at each face; see also the notes below

    returns:
       3 numpy float arrays of shape (nk + 1, nj, ni), (nk, nj + 1, ni), (nk, nj, ni + 1) being the
       neighbourly transmissibilities in K, J & I axes respectively

    notes:
       the 3 permeability arrays (and net to gross ratio if in use) must be identifiable in the property
       collection as they are used for the calculation;
       implicit units of measure of returned values will be m3.cP/(kPa.d) if grid crs length units are metres,
       bbl.cP/(psi.d) if length units are feet; the computation is compatible with the Nexus NEWTRAN formulation;
       values will be zero at pinchouts, and at column edges where there is a split pillar, even if there is
       juxtapostion of faces; the same is true of K gap faces (even where the gap is zero); NaNs in any of
       the feed properties also result in transmissibility values of zero;
       outer facing values will always be zero (included to be in accordance with RESQML faces properties);
       array caching in the grid object will only be used if realization is None; if a modifier mode of
       'faces multiplier' or 'faces per cell multiplier' is specified, properties will be searched for with
       local property kind 'transmissibility multiplier' and the appropriate indexable element (and optionally
       the direction facet in the case of 'faces multiplier'); the modifier mode of 'absolute' can be used to
       model the effect of faults and thin shales, tar mats etc. in a way which is independent of cell size;
       for 'aboslute' properties with indexable element of 'faces' (optionally directional) and local property kind
       'fault transmissibility' (or 'mat transmissibility') will be used; such absolute faces transmissibilities
       should have a value of np.inf or np.nan where no modification is required; note that this method is only
       dealing with logically neighbouring cells and will not compute values for faces with a split pillar,
       which should be handled elsewhere
    """

    # todo: improve handling of units: check uom for half cell transmissibility property and for absolute modifiers

    k_tr = j_tr = i_tr = None

    if realization is None:
        if hasattr(grid, 'array_k_transmissibility') and grid.array_k_transmissibility is not None:
            k_tr = grid.array_k_transmissibility
        if hasattr(grid, 'array_j_transmissibility') and grid.array_j_transmissibility is not None:
            j_tr = grid.array_j_transmissibility
        if hasattr(grid, 'array_i_transmissibility') and grid.array_i_transmissibility is not None:
            i_tr = grid.array_i_transmissibility

    if use_tr_properties and (k_tr is None or j_tr is None or i_tr is None):

        pc = grid.extract_property_collection()

        composite_tr = pc.single_array_ref(property_kind = 'transmissibility',
                                           realization = realization,
                                           continuous = True,
                                           indexable = 'faces',
                                           facet_type = 'none',
                                           facet = 'none')
        if k_tr is None:
            k_tr = pc.single_array_ref(property_kind = 'transmissibility',
                                       realization = realization,
                                       continuous = True,
                                       count = 1,
                                       indexable = 'faces',
                                       facet_type = 'direction',
                                       facet = 'K')
            if k_tr is None and composite_tr is not None:
                k_tr = composite_tr[:(grid.nk + 1) * grid.nj * grid.ni].reshape((grid.nk + 1, grid.nj, grid.ni))
            if k_tr is not None:
                assert k_tr.shape == (grid.nk + 1, grid.nj, grid.ni)
                if realization is None:
                    grid.array_k_transmissibility = k_tr

        if j_tr is None:
            j_tr = pc.single_array_ref(property_kind = 'transmissibility',
                                       realization = realization,
                                       continuous = True,
                                       count = 1,
                                       indexable = 'faces',
                                       facet_type = 'direction',
                                       facet = 'J')
            if j_tr is None and composite_tr is not None:
                offset = (grid.nk + 1) * grid.nj * grid.ni
                j_tr = composite_tr[offset:offset + grid.nk * (grid.nj + 1) * grid.ni].reshape(
                    (grid.nk, grid.nj + 1, grid.ni))
            if j_tr is not None:
                assert j_tr.shape == (grid.nk, grid.nj + 1, grid.ni)
                if realization is None:
                    grid.array_j_transmissibility = j_tr

        if i_tr is None:
            i_tr = pc.single_array_ref(property_kind = 'transmissibility',
                                       realization = realization,
                                       continuous = True,
                                       count = 1,
                                       indexable = 'faces',
                                       facet_type = 'direction',
                                       facet = 'I')
            if i_tr is None and composite_tr is not None:
                offset = (grid.nk + 1) * grid.nj * grid.ni + grid.nk * (grid.nj + 1) * grid.ni
                i_tr = composite_tr[offset:].reshape((grid.nk, grid.nj, grid.ni + 1))
            if i_tr is not None:
                assert i_tr.shape == (grid.nk, grid.nj, grid.ni + 1)
                if realization is None:
                    grid.array_i_transmissibility = i_tr

    if k_tr is None or j_tr is None or i_tr is None:

        half_t = grid.half_cell_transmissibility(use_property = use_tr_properties, realization = realization)

        if modifier_mode == 'faces per cell multiplier':
            half_t, pc = __faces_per_cell_multiplier(grid, half_t, pc, realization)

        if grid.has_split_coordinate_lines and (j_tr is None or i_tr is None):
            split_column_edges_j, split_column_edges_i = grid.split_column_faces()
        else:
            split_column_edges_j, split_column_edges_i = None, None

        np.seterr(divide = 'ignore')

        if k_tr is None:
            k_tr = __set_k_transmissibility(grid, half_t, k_tr, modifier_mode, pc, realization, tolerance)

        if j_tr is None:
            j_tr = __set_j_transmissibility(grid, half_t, j_tr, modifier_mode, pc, realization, split_column_edges_j,
                                            tolerance)

        if i_tr is None:
            i_tr = __set_i_transmissibility(grid, half_t, i_tr, modifier_mode, pc, realization, split_column_edges_i,
                                            tolerance)

        np.seterr(divide = 'warn')

    return k_tr, j_tr, i_tr


def __faces_per_cell_multiplier(grid, half_t, pc, realization):
    pc = grid.extract_property_collection()
    half_t_mult = pc.single_array_ref(property_kind = 'transmissibility multiplier',
                                      realization = realization,
                                      continuous = True,
                                      count = 1,
                                      indexable = 'faces per cell')
    if half_t_mult is None:
        log.warning('no faces per cell transmissibility multiplier found when calculating transmissibilities')
    else:
        log.debug('applying faces per cell transmissibility multipliers')
        half_t = np.where(np.isnan(half_t_mult), half_t, half_t * half_t_mult)
    return half_t, pc


def __set_i_transmissibility(grid, half_t, i_tr, modifier_mode, pc, realization, split_column_edges_i, tolerance):
    i_tr = np.zeros((grid.nk, grid.nj, grid.ni + 1))
    slice_a = half_t[:, :, :-1, 2, 1]  # note: internal faces only
    slice_b = half_t[:, :, 1:, 2, 0]
    internal_zero_mask = np.logical_or(np.logical_or(np.isnan(slice_a), slice_a < tolerance),
                                       np.logical_or(np.isnan(slice_b), slice_b < tolerance))
    if split_column_edges_i is not None:
        internal_zero_mask[:, split_column_edges_i] = True
    tr_mult = None
    if modifier_mode == 'faces multiplier':
        tr_mult = pc.single_array_ref(property_kind = 'transmissibility multiplier',
                                      realization = realization,
                                      facet_type = 'direction',
                                      facet = 'I',
                                      continuous = True,
                                      count = 1,
                                      indexable = 'faces')
        if tr_mult is None:
            tr_composite = pc.single_array_ref(property_kind = 'transmissibility multiplier',
                                               realization = realization,
                                               facet_type = 'none',
                                               facet = 'none',
                                               continuous = True,
                                               indexable = 'faces')
            if tr_composite is not None:
                offset = (grid.nk + 1) * grid.nj * grid.ni + grid.nk * (grid.nj + 1) * grid.ni
                tr_mult = tr_composite[offset:].reshape((grid.nk, grid.nj, grid.ni + 1))
        if tr_mult is None:
            log.warning('no I direction faces transmissibility multiplier found when calculating transmissibilities')
        else:
            assert tr_mult.shape == (grid.nk, grid.nj, grid.ni + 1)
            internal_zero_mask = np.logical_or(internal_zero_mask, np.isnan(tr_mult[:, :, 1:-1]))
    if tr_mult is None:
        tr_mult = 1.0
    tr_abs_r = 0.0
    if modifier_mode == 'absolute':
        tr_abs = pc.single_array_ref(property_kind = 'fault transmissibility',
                                     realization = realization,
                                     facet_type = 'direction',
                                     facet = 'I',
                                     continuous = True,
                                     count = 1,
                                     indexable = 'faces')
        if tr_abs is not None:
            log.debug('applying absolute I face transmissibility modification')
            assert tr_abs.shape == (grid.nk, grid.nj, grid.ni + 1)
            internal_zero_mask = np.logical_or(internal_zero_mask, tr_abs[:, :, 1:-1] <= 0.0)
            tr_abs_r = np.where(np.logical_or(np.isinf(tr_abs), np.isnan(tr_abs)), 0.0, 1.0 / tr_abs)
    i_tr[:, :, 1:-1] = np.where(internal_zero_mask, 0.0, tr_mult / ((1.0 / slice_a) + tr_abs_r + (1.0 / slice_b)))
    return i_tr


def __set_j_transmissibility(grid, half_t, j_tr, modifier_mode, pc, realization, split_column_edges_j, tolerance):
    j_tr = np.zeros((grid.nk, grid.nj + 1, grid.ni))
    slice_a = half_t[:, :-1, :, 1, 1]  # note: internal faces only
    slice_b = half_t[:, 1:, :, 1, 0]
    internal_zero_mask = np.logical_or(np.logical_or(np.isnan(slice_a), slice_a < tolerance),
                                       np.logical_or(np.isnan(slice_b), slice_b < tolerance))
    if split_column_edges_j is not None:
        internal_zero_mask[:, split_column_edges_j] = True
    tr_mult = None
    if modifier_mode == 'faces multiplier':
        tr_mult = pc.single_array_ref(property_kind = 'transmissibility multiplier',
                                      realization = realization,
                                      facet_type = 'direction',
                                      facet = 'J',
                                      continuous = True,
                                      count = 1,
                                      indexable = 'faces')
        if tr_mult is None:
            tr_composite = pc.single_array_ref(property_kind = 'transmissibility multiplier',
                                               realization = realization,
                                               facet_type = 'none',
                                               facet = 'none',
                                               continuous = True,
                                               indexable = 'faces')
            if tr_composite is not None:
                offset = (grid.nk + 1) * grid.nj * grid.ni
                tr_mult = tr_composite[offset:offset + grid.nk * (grid.nj + 1) * grid.ni].reshape(
                    (grid.nk, grid.nj + 1, grid.ni))
        if tr_mult is None:
            log.warning('no J direction faces transmissibility multiplier found when calculating transmissibilities')
        else:
            assert tr_mult.shape == (grid.nk, grid.nj + 1, grid.ni)
            internal_zero_mask = np.logical_or(internal_zero_mask, np.isnan(tr_mult[:, 1:-1, :]))
    if tr_mult is None:
        tr_mult = 1.0
    tr_abs_r = 0.0
    if modifier_mode == 'absolute':
        tr_abs = pc.single_array_ref(property_kind = 'fault transmissibility',
                                     realization = realization,
                                     facet_type = 'direction',
                                     facet = 'J',
                                     continuous = True,
                                     count = 1,
                                     indexable = 'faces')
        if tr_abs is not None:
            log.debug('applying absolute J face transmissibility modification')
            assert tr_abs.shape == (grid.nk, grid.nj + 1, grid.ni)
            internal_zero_mask = np.logical_or(internal_zero_mask, tr_abs[:, 1:-1, :] <= 0.0)
            tr_abs_r = np.where(np.logical_or(np.isinf(tr_abs), np.isnan(tr_abs)), 0.0, 1.0 / tr_abs)
    j_tr[:, 1:-1, :] = np.where(internal_zero_mask, 0.0, tr_mult / ((1.0 / slice_a) + tr_abs_r + (1.0 / slice_b)))
    if realization is None:
        grid.array_j_transmissibility = j_tr
    return j_tr


def __set_k_transmissibility(grid, half_t, k_tr, modifier_mode, pc, realization, tolerance):
    k_tr = np.zeros((grid.nk + 1, grid.nj, grid.ni))
    slice_a = half_t[:-1, :, :, 0, 1]  # note: internal faces only
    slice_b = half_t[1:, :, :, 0, 0]
    internal_zero_mask = np.logical_or(np.logical_or(np.isnan(slice_a), slice_a < tolerance),
                                       np.logical_or(np.isnan(slice_b), slice_b < tolerance))
    if grid.k_gaps:  # todo: scan K gaps for zero thickness gaps and allow transmission there
        internal_zero_mask[grid.k_gap_after_array, :, :] = True
    tr_mult = None
    if modifier_mode == 'faces multiplier':
        tr_mult = pc.single_array_ref(property_kind = 'transmissibility multiplier',
                                      realization = realization,
                                      facet_type = 'direction',
                                      facet = 'K',
                                      continuous = True,
                                      count = 1,
                                      indexable = 'faces')
        if tr_mult is None:
            tr_composite = pc.single_array_ref(property_kind = 'transmissibility multiplier',
                                               realization = realization,
                                               facet_type = 'none',
                                               facet = 'none',
                                               continuous = True,
                                               indexable = 'faces')
            if tr_composite is not None:
                size = (grid.nk + 1) * grid.nj * grid.ni
                tr_mult = tr_composite[:size].reshape((grid.nk + 1, grid.nj, grid.ni))
        if tr_mult is not None:
            assert tr_mult.shape == (grid.nk + 1, grid.nj, grid.ni)
            internal_zero_mask = np.logical_or(internal_zero_mask, np.isnan(tr_mult[1:-1, :, :]))
    if tr_mult is None:
        tr_mult = 1.0
    tr_abs_r = 0.0
    if modifier_mode == 'absolute':
        tr_abs = pc.single_array_ref(property_kind = 'mat transmissibility',
                                     realization = realization,
                                     facet_type = 'direction',
                                     facet = 'K',
                                     continuous = True,
                                     count = 1,
                                     indexable = 'faces')
        if tr_abs is None:
            tr_abs = pc.single_array_ref(property_kind = 'mat transmissibility',
                                         realization = realization,
                                         continuous = True,
                                         count = 1,
                                         indexable = 'faces')
        if tr_abs is None:
            tr_abs = pc.single_array_ref(property_kind = 'fault transmissibility',
                                         realization = realization,
                                         facet_type = 'direction',
                                         facet = 'K',
                                         continuous = True,
                                         count = 1,
                                         indexable = 'faces')
        if tr_abs is not None:
            log.debug('applying absolute K face transmissibility modification')
            assert tr_abs.shape == (grid.nk + 1, grid.nj, grid.ni)
            internal_zero_mask = np.logical_or(internal_zero_mask, tr_abs[1:-1, :, :] <= 0.0)
            tr_abs_r = np.where(np.logical_or(np.isinf(tr_abs), np.isnan(tr_abs)), 0.0, 1.0 / tr_abs)
    k_tr[1:-1, :, :] = np.where(internal_zero_mask, 0.0, tr_mult / ((1.0 / slice_a) + tr_abs_r + (1.0 / slice_b)))
    if realization is None:
        grid.array_k_transmissibility = k_tr
    return k_tr


def half_cell_transmissibility(grid, use_property = True, realization = None, tolerance = 1.0e-6):
    """Returns (and caches if realization is None) half cell transmissibilities for this grid.

    arguments:
       use_property (boolean, default True): if True, the grid's property collection is inspected for
          a possible half cell transmissibility array and if found, it is used instead of calculation
       realization (int, optional) if present, only a property with this realization number will be used
       tolerance (float, default 1.0e-6): minimum half axis length below which the transmissibility
          will be deemed uncomputable (for the axis in question); NaN values will be returned (not Inf);
          units are implicitly those of the grid's crs length units

    returns:
       numpy float array of shape (nk, nj, ni, 3, 2) where the 3 covers K,J,I and the 2 covers the
          face polarity: - (0) and + (1); units will depend on the length units of the coordinate reference
          system for the grid; the units will be m3.cP/(kPa.d) or bbl.cP/(psi.d) for grid length units of m
          and ft respectively

    notes:
       the returned array is in the logical resqpy arrangement; it must be discombobulated before being
       added as a property; this method does not write to hdf5, nor create a new property or xml;
       if realization is None, a grid attribute cached array will be used; tolerance will only be
       used if the half cell transmissibilities are actually computed
    """

    # todo: allow passing of property uuids for ntg, k_k, j, i

    if realization is None and hasattr(grid, 'array_half_cell_t'):
        return grid.array_half_cell_t

    half_t = None

    if use_property:
        pc = grid.property_collection
        half_t_resqml = pc.single_array_ref(property_kind = 'transmissibility',
                                            realization = realization,
                                            continuous = True,
                                            count = 1,
                                            indexable = 'faces per cell')
        if half_t_resqml:
            assert half_t_resqml.shape == (grid.nk, grid.nj, grid.ni, 6)
            half_t = pc.combobulate(half_t_resqml)

    if half_t is None:
        # note: properties must be identifiable in property_collection
        half_t = rqtr.half_cell_t(grid, realization = realization)

    if realization is None:
        grid.array_half_cell_t = half_t

    return half_t



"""Submodule containing the RegularGrid class."""

import logging

log = logging.getLogger(__name__)

import math as maths
import numpy as np

import resqpy.crs as rqc
import resqpy.grid
import resqpy.olio.transmission as rqtr
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vec
import resqpy.olio.xml_et as rqet
import resqpy.property as rprop
from resqpy.olio.xml_namespaces import curly_namespace as ns

import resqpy.grid._grid as grr_g
import resqpy.grid._grid_types as grr_gt
import resqpy.grid._create_grid_xml as grr_xml

# for a regular grid, the ...is_defined xml is created as a constant bool array, if the following set True
always_write_pillar_geometry_is_defined = True  # required True for Fesapi interoperability (xsd compliance)
always_write_cell_geometry_is_defined = False


class RegularGrid(grr_g.Grid):
    """Class for completely regular block grids, usually aligned with xyz axes."""

    # WIP: use RESQML lattice like geometry specification

    def __init__(self,
                 parent_model,
                 uuid = None,
                 extent_kji = None,
                 dxyz = None,
                 dxyz_dkji = None,
                 origin = (0.0, 0.0, 0.0),
                 crs_uuid = None,
                 use_vertical = False,
                 mesh = None,
                 mesh_dz_dk = 1.0,
                 set_points_cached = False,
                 as_irregular_grid = False,
                 find_properties = True,
                 title = None,
                 originator = None,
                 extra_metadata = {}):
        """Creates a regular grid object based on dxyz, or derived from a Mesh object.

        arguments:
           parent_model (model.Model object): the model to which the new grid will be assigned
           uuid (optional): the uuid for an existing grid part; if present, the RegularGrid object is
              based on existing data or a mix of that data and other arguments where present
           extent_kji (triple positive integers, optional): the number of cells in the grid (nk, nj, ni);
              required unless uuid is present
           dxyz (triple float, optional): use when the I,J,K axes align with the x,y,z axes (possible with inverted
              directions); the size of each cell (dx, dy, dz); values may be negative
           dxyz_dkji (numpy float array of shape (3, 3), optional): how x,y,z values increase with each step in each
              direction K,J,I; first index is KJI, second index is xyz; only one of dxyz, dxyz_dkji and mesh should be
              present; NB axis ordering is different to that used in Mesh class for dxyz_dij
           origin (triple float, default (0.0, 0.0, 0.0)): the location in the local coordinate space of the crs of
              the 'first' corner point of the grid
           crs_uuid (uuid.UUID, optional): the uuid of the coordinate reference system for the grid
           use_vertical (boolean, default False): if True and the pillars of the regular grid are vertical then a
              pillar shape of 'vertical' is used; if False (or the pillars are not vertical), then a pillar shape of
              'straight' is used
           mesh (surface.Mesh object, optional): if present, the I,J layout of the grid is based on the mesh, which
              must be regular, and the K cell size is given by the mesh_dz_dk argument; if present, then dxyz and
              dxyz_dkji must be None
           mesh_dz_dk (float, default 1.0): the size of cells in the K axis, which is aligned with the z axis, when
              starting from a mesh; ignored if mesh is None
           set_points_cached (boolean, default False): if True, an explicit geometry is created for the regular grid
              in the form of the cached points array; will be treated as True if as_irregular_grid is True
           as_irregular_grid (boolean, default False): if True, the grid is setup such that it will appear as a Grid
              object when next loaded from disc
           find_properties (boolean, default True): if True and uuid is not None, a grid property collection is
              instantiated as an attribute, holding properties for which this grid is the supporting representation
           title (str, optional): citation title for new grid; ignored if loading from xml
           originator (str, optional): name of person creating the grid; defaults to login id;
              ignored if loading from xml
           extra_metadata (dict, optional): dictionary of extra metadata items to add to the grid;
              ignored if loading from xml

        returns:
           a newly created RegularGrid object with inheritance from the Grid class

        notes:
           the RESQML standard allows for regular grid geometry pillars to be stored as parametric lines
           but that is not yet supported by this code base; however, constant dx, dy, dz arrays are supported;
           alternatively, regular meshes (Grid2d) may be stored in parameterized form and used to generate a
           regular grid here;
           if uuid, dxyz, dxyz_dkji and mesh arguments are all None then unit cube cells aligned with
           the x,y,z axes will be generated;
           to store the geometry explicitly set as_irregular_grid True and use the following methods:
           make_regular_points_cached(), write_hdf5(), create_xml(..., write_geometry = True);
           otherwise, avoid write_hdf5() and call create_xml(..., write_geometry = False);
           if geometry is not stored explicitly, the uuid of the crs is stored as extra metadata
           if origin is not triple zero, a new crs will be created with the origin moved appropriately

        :meta common:
        """

        if as_irregular_grid:
            set_points_cached = True
            self.is_aligned = False
        else:
            self.is_aligned = None  #: boolean indicating alignment of IJK axes with +/- xyz respectively

        if uuid is None:

            assert extent_kji is not None and len(extent_kji) == 3
            super().__init__(parent_model, title = title, originator = originator, extra_metadata = extra_metadata)
            self.grid_representation = 'IjkGrid' if as_irregular_grid else 'IjkBlockGrid'
            self.extent_kji = np.array(extent_kji).copy()
            self.nk, self.nj, self.ni = self.extent_kji
            self.k_direction_is_down = True  # assumed direction
            self.grid_is_right_handed = False  # todo: work it out from dxyz_dkji and crs xyz handedness
            self.has_split_coordinate_lines = False
            self.k_gaps = None
            self.k_gap_after_array = None
            self.k_raw_index_array = None
            self.inactive = None
            self.all_inactive = None

            if mesh is not None:
                assert mesh.flavour == 'regular'
                assert dxyz is None and dxyz_dkji is None
                origin = mesh.regular_origin
                dxyz_dkji = np.empty((3, 3))
                dxyz_dkji[0, :] = mesh_dz_dk
                dxyz_dkji[1, :] = mesh.regular_dxyz_dij[1]  # J axis
                dxyz_dkji[2, :] = mesh.regular_dxyz_dij[0]  # I axis
                if crs_uuid is None:
                    crs_uuid = mesh.crs_uuid
                else:
                    assert bu.matching_uuids(crs_uuid, mesh.crs_uuid)

            assert dxyz is None or dxyz_dkji is None
            if dxyz is None and dxyz_dkji is None:
                dxyz = (1.0, 1.0, 1.0)
            if dxyz_dkji is None:
                dxyz_dkji = np.array([[0.0, 0.0, dxyz[2]], [0.0, dxyz[1], 0.0], [dxyz[0], 0.0, 0.0]])
            self.block_origin = np.array(origin).copy() if uuid is None else np.zeros(3)
            self.block_dxyz_dkji = np.array(dxyz_dkji).copy()
            if self.is_aligned is None:
                self._set_is_aligned()
            if use_vertical and dxyz_dkji[0][0] == 0.0 and dxyz_dkji[0][1] == 0.0:  # ie. no x,y change with k
                self.pillar_shape = 'vertical'
            else:
                self.pillar_shape = 'straight'

            new_crs = None
            shift_origin = np.any(origin != 0.0) and uuid is None and not as_irregular_grid
            if crs_uuid is None and self.extra_metadata is not None:
                crs_uuid = bu.uuid_from_string(self.extra_metadata.get('crs uuid'))
                shift_origin = shift_origin and crs_uuid is None
            if crs_uuid is None:
                crs_uuid = parent_model.crs_uuid
            if crs_uuid is None:
                new_crs = rqc.Crs(parent_model, x_offset = origin[0], y_offset = origin[1], z_offset = origin[2])
                shift_origin = False
                new_crs.create_xml(reuse = True)
                crs_uuid = new_crs.uuid
            if shift_origin:
                new_crs = rqc.Crs(parent_model, uuid = crs_uuid)
                new_crs.uuid = bu.new_uuid()
                new_crs.x_offset += origin[0]
                new_crs.y_offset += origin[1]
                new_crs.z_offset += origin[2]
                new_crs.create_xml(reuse = True)
                crs_uuid = new_crs.uuid
            self.crs_uuid = crs_uuid
            self.crs = rqc.Crs(parent_model, uuid = crs_uuid) if new_crs is None else new_crs

        else:

            assert bu.is_uuid(uuid)
            assert grr_gt.is_regular_grid(parent_model.root_for_uuid(uuid))
            super().__init__(parent_model,
                             uuid = uuid,
                             find_properties = find_properties,
                             geometry_required = False,
                             title = title,
                             originator = originator,
                             extra_metadata = extra_metadata)

        self.geometry_defined_for_all_cells_cached = True
        self.geometry_defined_for_all_pillars_cached = True
        # self.array_cell_geometry_is_defined = np.full(tuple(self.extent_kji), True, dtype = bool)
        self.array_cell_geometry_is_defined = None

        if set_points_cached:
            self.make_regular_points_cached()

        if self.uuid is None:
            self.uuid = bu.new_uuid()

    def _load_regular_grid_from_xml(self):
        """Load from xml, called from super() init."""

        def get_point3d(node):
            assert node is not None and rqet.node_type(node) == 'Point3d'
            xyz = []
            for c in '123':
                xyz.append(rqet.find_tag_float(node, f'Coordinate{c}'))
                assert xyz[-1] is not None
            return tuple(xyz)

        self.grid_representation = 'IjkBlockGrid'
        self.is_aligned = True  # currently only aligned regular grids supported for lattice mode
        self.geometry_defined_for_all_pillars_cached = True
        self.geometry_defined_for_all_cells_cached = True
        self.pillar_shape = 'vertical'
        self.has_split_coordinate_lines = False
        self.block_origin = np.zeros(3, dtype = float)
        self.block_dxyz_dkji = np.zeros((3, 3), dtype = float)
        if self.geometry_root is None:
            log.debug('loading from xml without geometry node')
            # find cell length properties and populate dxyz from those values
            pc = self.extract_property_collection()
            assert pc is not None
            dxi_part = pc.singleton(property_kind = 'cell length',
                                    facet_type = 'direction',
                                    facet = 'I',
                                    indexable = 'cells',
                                    const_value = '*')
            dyj_part = pc.singleton(property_kind = 'cell length',
                                    facet_type = 'direction',
                                    facet = 'J',
                                    indexable = 'cells',
                                    const_value = '*')
            dzk_part = pc.singleton(property_kind = 'cell length',
                                    facet_type = 'direction',
                                    facet = 'K',
                                    indexable = 'cells',
                                    const_value = '*')
            assert dxi_part is not None and dyj_part is not None and dzk_part is not None
            dxi = self.property_collection.constant_value_for_part(dxi_part)
            dyj = self.property_collection.constant_value_for_part(dyj_part)
            dzk = self.property_collection.constant_value_for_part(dzk_part)
            assert dxi is not None and dyj is not None and dzk is not None
            self.block_dxyz_dkji[2, 0] = dxi
            self.block_dxyz_dkji[1, 1] = dyj
            self.block_dxyz_dkji[0, 2] = dzk
        else:
            log.debug('loading from xml with lattice geometry node')
            p_node = rqet.find_tag(self.geometry_root, 'Points')
            assert p_node is not None and rqet.node_type(p_node) == 'Point3dLatticeArray'
            assert rqet.find_tag_bool(p_node, 'AllDimensionsAreOrthogonal',
                                      must_exist = True), 'only aligned regular grids catered for'
            origin_node = rqet.find_tag(p_node, 'Origin')
            assert origin_node is not None, 'origin missing in xml for regular mesh (lattice)'
            self.block_origin[:] = get_point3d(origin_node)
            # following constraint is for aligned regular grids
            assert np.all(self.block_origin == 0.0),  \
                   'aligned regular grid origin must be coincident with local crs origin'
            offset_nodes = rqet.list_of_tag(p_node, 'Offset')
            assert len(offset_nodes) == 3, 'missing (or too many) offset nodes in xml for regular grid (lattice)'
            for axis, main_o_node in enumerate(offset_nodes):  # order K, J, I; ie. z, y, x
                d_xyz = get_point3d(rqet.find_tag(main_o_node, 'Offset'))
                self.block_dxyz_dkji[axis] = d_xyz
                # following constraint is for aligned regular grids
                for ax in range(3):
                    # note: these constraints allow for a negative cell length, ie. xyz axis reverse of IJK
                    is_non_zero = not maths.isclose(d_xyz[2 - ax], 0.0)
                    assert (ax == axis) == is_non_zero, 'unaligned lattice for regular grid not supported'
                spacing_node = rqet.find_tag(main_o_node, 'Spacing')
                assert spacing_node is not None
                spacing_value = rqet.find_tag_float(spacing_node, 'Value')
                assert spacing_value is not None and maths.isclose(spacing_value, 1.0)
                spacing_count = rqet.find_tag_int(spacing_node, 'Count')
                assert spacing_count is not None and spacing_count == self.extent_kji[axis]
            # log.debug(f'block dxyz_dkji from lattice: {self.block_dxyz_dkji}')

    def cell_geometry_is_defined(self, cell_kji0 = None, cell_geometry_is_defined_root = None, cache_array = True):
        """Override of Grid method indicating whether a cell has defined geometry, here always True."""

        return True

    def pillar_geometry_is_defined(self, pillar_ji0 = None, pillar_geometry_is_defined_root = None, cache_array = True):
        """Override of Grid method indicating whether a pillar has defined geometry, here always True."""

        return True

    def make_regular_points_cached(self, apply_origin_offset = True):
        """Set up the cached points array as an explicit representation of the regular grid geometry.

        arguments:
           apply_origin_offset (boolean, default True): if True, this method includes the regular grid
           origin in the calculated points data

        note:
           if apply_origin_offset is True, the related crs must not store the origin as offset data
        """

        if hasattr(self, 'points_cached') and self.points_cached is not None:
            return
        self.points_cached = np.zeros((self.nk + 1, self.nj + 1, self.ni + 1, 3))
        # todo: replace for loops with linspace
        for k in range(self.nk):
            self.points_cached[k + 1, 0, 0] = self.points_cached[k, 0, 0] + self.block_dxyz_dkji[0]
        for j in range(self.nj):
            self.points_cached[:, j + 1, 0] = self.points_cached[:, j, 0] + self.block_dxyz_dkji[1]
        for i in range(self.ni):
            self.points_cached[:, :, i + 1] = self.points_cached[:, :, i] + self.block_dxyz_dkji[2]
        if apply_origin_offset:
            self.points_cached[:, :, :] += self.block_origin

    def axial_lengths_kji(self):
        """Returns a triple float being lengths of primary axes (K, J, I) for each cell."""

        return vec.naive_lengths(self.block_dxyz_dkji)

    # override of Grid methods

    def point_raw(self, index = None, points_root = None, cache_array = True):
        """Returns element from points data, indexed as corner point (k0, j0, i0).

        Can optionally be used to cache points data.

        arguments:
           index (3 integers, optional): if not None, the index into the raw points data for the point of interest
           points_root (ignored)
           cache_array (boolean, default True): if True, the raw points data is cached in memory as a side effect

        returns:
           (x, y, z) of selected point as a 3 element numpy vector, or None if index is None

        notes:
           this function is typically called either to cache the points data in memory, or to fetch the coordinates of
           a single corner point;
           the index should be a triple kji0 with axes ranging over the shared corners nk+1, nj+1, ni+1
        """

        assert cache_array or index is not None

        if cache_array:
            self.make_regular_points_cached()
            if index is None:
                return None
            return self.points_cached[tuple(index)]

        return self.block_origin + np.sum(np.repeat(np.array(index).reshape(
            (3, 1)), 3, axis = -1) * self.block_dxyz_dkji,
                                          axis = 0)

    def half_cell_transmissibility(self, use_property = None, realization = None, tolerance = None):
        """Returns (and caches if realization is None) half cell transmissibilities for this regular grid.

        arguments:
           use_property (ignored)
           realization (int, optional) if present, only a property with this realization number will be used
           tolerance (ignored)

        returns:
           numpy float array of shape (nk, nj, ni, 3, 2) where the 3 covers K,J,I and the 2 covers the
              face polarity: - (0) and + (1); units will depend on the length units of the coordinate reference
              system for the grid; the units will be m3.cP/(kPa.d) or bbl.cP/(psi.d) for grid length units of m
              and ft respectively

        notes:
           the values for - and + polarity will always be equal, the data is duplicated for compatibility with
           parent Grid class;
           the returned array is in the logical resqpy arrangement; it must be discombobulated before being
           added as a property; this method does not write to hdf5, nor create a new property or xml;
           if realization is None, a grid attribute cached array will be used
        """

        # todo: allow passing of property uuids for ntg, k_k, j, i

        if realization is None and hasattr(self, 'array_half_cell_t'):
            return self.array_half_cell_t

        half_t = rqtr.half_cell_t(
            self, realization = realization)  # note: properties must be identifiable in property_collection

        # introduce facial polarity axis for compatibility with parent Grid class
        assert half_t.ndim == 4
        half_t = np.expand_dims(half_t, -1)
        half_t = np.repeat(half_t, 2, axis = -1)

        if realization is None:
            self.array_half_cell_t = half_t

        return half_t

    def corner_points(self, cell_kji0 = None, points_root = None, cache_resqml_array = None, cache_cp_array = False):
        """Returns a numpy array of corner points for a single cell or the whole grid.

        arguments:
            cell_kji0 (triple float, optional): if present, index of cell for which corner points are required; if None,
                an array holding corner points for all cells is returned
            points_root (ignored): not used; exists for signature compatibility with parent class method
            cache_resqml_array (ignored): not used; exists for signature compatibility with parent class method
            cache_cp_array (bool, default False): if True (or cell_kji0 is None), the full corner point array is
                cached as an attribute of the grid

        notes:
           if cell_kji0 is not None, a 4D array of shape (2, 2, 2, 3) holding single cell corner points in logical order
           [kp, jp, ip, xyz] is returned; if cell_kji0 is None, a pagoda style 7D array [k, j, i, kp, jp, ip, xyz] is
           cached and returned;
           the ordering of the corner points is in the logical order, which is not the same as that used by Nexus CORP data;
           olio.grid_functions.resequence_nexus_corp() can be used to switch back and forth between this pagoda ordering
           and Nexus corp ordering;
           this is the usual way to access full corner points for cells where working with native resqml data is undesirable;
           this method returns coordinates in the local crs space

        :meta common:
        """

        if cell_kji0 is None:
            cache_cp_array = True
        if hasattr(self, 'array_corner_points'):
            if cell_kji0 is None:
                return self.array_corner_points
            return self.array_corner_points[tuple(cell_kji0)]
        if not self.is_aligned:
            return super().corner_points(cell_kji0 = cell_kji0,
                                         points_root = points_root,
                                         cache_resqml_array = cache_resqml_array,
                                         cache_cp_array = cache_cp_array)

        if cache_cp_array:
            acp = np.zeros((self.nk, self.nj, self.ni, 2, 2, 2, 3), dtype = float)
            # set x coordinates
            acp[:, :, :, 0, 0, 0, 0] = np.linspace(0.0,
                                                   self.block_dxyz_dkji[2, 0] * self.ni,
                                                   num = self.ni,
                                                   endpoint = False).reshape((1, 1, self.ni))
            acp[:, :, :, 0, 0, 1, 0] = acp[:, :, :, 0, 0, 0, 0] + self.block_dxyz_dkji[2, 0]
            acp[:, :, :, 0, 1, :, 0] = acp[:, :, :, 0, 0, :, 0]
            acp[:, :, :, 1, :, :, 0] = acp[:, :, :, 0, :, :, 0]
            # set y coordinates
            acp[:, :, :, 0, 0, 0, 1] = np.linspace(0.0,
                                                   self.block_dxyz_dkji[1, 1] * self.nj,
                                                   num = self.nj,
                                                   endpoint = False).reshape((1, self.nj, 1))
            acp[:, :, :, 0, 1, 0, 1] = acp[:, :, :, 0, 0, 0, 1] + self.block_dxyz_dkji[1, 1]
            acp[:, :, :, 0, :, 1, 1] = acp[:, :, :, 0, :, 0, 1]
            acp[:, :, :, 1, :, :, 1] = acp[:, :, :, 0, :, :, 1]
            # set z coordinates
            acp[:, :, :, 0, 0, 0, 2] = np.linspace(0.0,
                                                   self.block_dxyz_dkji[0, 2] * self.nk,
                                                   num = self.nk,
                                                   endpoint = False).reshape((self.nk, 1, 1))
            acp[:, :, :, 1, 0, 0, 2] = acp[:, :, :, 0, 0, 0, 2] + self.block_dxyz_dkji[0, 2]
            acp[:, :, :, :, 0, 1, 2] = acp[:, :, :, :, 0, 0, 2]
            acp[:, :, :, :, 1, :, 2] = acp[:, :, :, :, 0, :, 2]
            # translate by regular grid origin
            acp[:] += np.array(self.block_origin, dtype = float).reshape((1, 1, 1, 1, 1, 1, 3))
            self.array_corner_points = acp
            if cell_kji0 is None:
                return acp
            return acp[tuple(cell_kji0)]

        # setup corner point array for a single cell only
        cp = np.zeros((2, 2, 2, 3), dtype = float)
        # set x coordinates
        cp[0, 0, 0, 0] = self.block_dxyz_dkji[2, 0] * cell_kji0[2]
        cp[0, 0, 1, 0] = cp[0, 0, 0, 0] + self.block_dxyz_dkji[2, 0]
        cp[0, 1, :, 0] = cp[0, 0, :, 0]
        cp[1, :, :, 0] = cp[0, :, :, 0]
        # set y coordinates
        cp[0, 0, 0, 1] = self.block_dxyz_dkji[1, 1] * cell_kji0[1]
        cp[0, 1, 0, 1] = cp[0, 0, 0, 1] + self.block_dxyz_dkji[1, 1]
        cp[0, :, 1, 1] = cp[0, :, 0, 1]
        cp[1, :, :, 1] = cp[0, :, :, 1]
        # set z coordinates
        cp[0, 0, 0, 2] = self.block_dxyz_dkji[0, 2] * cell_kji0[0]
        cp[1, 0, 0, 2] = cp[0, 0, 0, 2] + self.block_dxyz_dkji[0, 2]
        cp[:, 0, 1, 2] = cp[:, 0, 0, 2]
        cp[:, 1, :, 2] = cp[:, 0, :, 2]
        # translate by regular grid origin
        cp[:] += np.array(self.block_origin, dtype = float).reshape((1, 1, 1, 3))
        return cp

    def centre_point(self, cell_kji0 = None, cache_centre_array = False, use_origin = True):
        """Returns centre point of a cell or array of centre points of all cells.

        arguments:
           cell_kji0 (optional): if present, the (k, j, i) indices of the individual cell for which the
              centre point is required; zero based indexing
           cache_centre_array (bool, default False): If True, or cell_kji0 is None, an array of centre points
              is generated and added as an attribute of the grid, with attribute name array_centre_point
           use_origin (bool, default True): if True, the x, y & z offsets (local origin) are added to the
              computed cell centre points

        returns:
           (x, y, z) 3 element numpy array of floats holding centre point of cell;
           or numpy 3+1D array if cell_kji0 is None

        note:
           resulting coordinates are in the same (local) crs as the grid points if use_origin is True
        """

        if cell_kji0 is None:
            cache_centre_array = True

        if cache_centre_array and (not hasattr(self, 'array_centre_point') or self.array_centre_point is None or
                                   not use_origin):
            centres = np.zeros((self.nk, self.nj, self.ni, 3))
            if self.is_aligned:
                centres[:, :, :, 0] = np.linspace(0.0,
                                                  self.block_dxyz_dkji[2, 0] * self.ni,
                                                  num = self.ni,
                                                  endpoint = False).reshape((1, 1, self.ni))
                centres[:, :, :, 1] = np.linspace(0.0,
                                                  self.block_dxyz_dkji[1, 1] * self.nj,
                                                  num = self.nj,
                                                  endpoint = False).reshape((1, self.nj, 1))
                centres[:, :, :, 2] = np.linspace(0.0,
                                                  self.block_dxyz_dkji[0, 2] * self.nk,
                                                  num = self.nk,
                                                  endpoint = False).reshape((self.nk, 1, 1))
            else:
                # todo: replace for loops with linspace
                for k in range(self.nk - 1):
                    centres[k + 1, 0, 0] = centres[k, 0, 0] + self.block_dxyz_dkji[0]
                for j in range(self.nj - 1):
                    centres[:, j + 1, 0] = centres[:, j, 0] + self.block_dxyz_dkji[1]
                for i in range(self.ni - 1):
                    centres[:, :, i + 1] = centres[:, :, i] + self.block_dxyz_dkji[2]
            centres += 0.5 * np.sum(self.block_dxyz_dkji, axis = 0)
            if use_origin:
                centres += self.block_origin
                self.array_centre_point = centres
        else:
            centres = None

        if cell_kji0 is not None:
            if centres is not None:
                return centres[tuple(cell_kji0)]
            if hasattr(self, 'array_centre_point') and self.array_centre_point is not None and use_origin:
                return self.array_centre_point[tuple(cell_kji0)]
            float_kji0 = np.array(cell_kji0, dtype = float) + 0.5
            centre = np.sum(self.block_dxyz_dkji * np.expand_dims(float_kji0, axis = -1).repeat(3, axis = -1), axis = 0)
            if use_origin:
                centre += self.block_origin
            return centre

        if centres is not None:
            return centres

        return self.array_centre_point

    def aligned_column_centres(self):
        """For an aligned grid, returns an array of column centres in xy, of shape (nj, ni, 2)."""

        assert self.is_aligned
        centres = np.zeros((self.nj, self.ni, 2))
        centres[:, :, 0] = np.linspace(0.0, self.block_dxyz_dkji[2, 0] * self.ni, num = self.ni,
                                       endpoint = False).reshape((1, self.ni))
        centres[:, :, 1] = np.linspace(0.0, self.block_dxyz_dkji[1, 1] * self.nj, num = self.nj,
                                       endpoint = False).reshape((self.nj, 1))
        centres += self.block_origin[:2] + 0.5 * np.sum(self.block_dxyz_dkji[1:, :2], axis = 0)
        return centres

    def volume(self, cell_kji0 = None, required_uom = None):
        """Returns bulk rock volume of cell or numpy array of bulk rock volumes for all cells.

        arguments:
           cell_kji0 (optional): if present, the (k, j, i) indices of the individual cell for which the
                                 volume is required; zero based indexing
           required_uom (str, optional): if present, the RESQML unit of measure (for quantity volume) that
                                 the volumes will be returned in; if None, the grid's CRS z units cubed
                                 will be used

        returns:
           float, being the volume of cell identified by cell_kji0;
           or numpy float array of shape (nk, nj, ni) if cell_kji0 is None

        notes:
           the function can be used to find the volume of a single cell, or all cells;
           the method currently assumes that the primary i, j, k axes are mutually orthogonal
        """

        required_uom = self.get_volume_uom(required_uom)
        conversion_factor = self.get_volume_conversion_factor(required_uom)
        vol = np.prod(vec.naive_lengths(self.block_dxyz_dkji))
        if conversion_factor is not None:
            vol *= conversion_factor
        if cell_kji0 is not None:
            return vol
        return np.full((self.nk, self.nj, self.ni), vol)

    def thickness(self, cell_kji0 = None, **kwargs):
        """Returns cell thickness (K axial length) for a single cell or full array.

        arguments:
           cell_kji0 (triple int, optional): if present, the thickness for a single cell is returned;
              if None, an array is returned
           all other arguments ignored; present for compatibility with same method in Grid()

        returns:
           float, or numpy float array filled with a constant
        """

        thick = self.axial_lengths_kji()[0]
        if cell_kji0 is not None:
            return thick
        return np.full((self.nk, self.nj, self.ni), thick, dtype = float)

    def pinched_out(self, cell_kji0 = None, **kwargs):
        """Returns pinched out boolean (always False) for a single cell or full array.

        arguments:
           cell_kji0 (triple int, optional): if present, the pinched out flag for a single cell is returned;
              if None, an array is returned
           all other arguments ignored; present for compatibility with same method in Grid()

        returns:
           False, or numpy array filled with False
        """

        if cell_kji0 is not None:
            return False
        return np.full((self.nk, self.nj, self.ni), False, dtype = bool)

    def actual_pillar_shape(self, patch_metadata = False, tolerance = 0.001):
        """Returns actual shape of pillars.

        arguments:
           patch_metadata (boolean, default False): if True, the actual shape replaces whatever was in the metadata
           tolerance (float, ignored)

        returns:
           string: 'vertical', 'straight' or 'curved'

        note:
           setting patch_metadata True will affect the attribute in this Grid object; however, it will not be
           preserved unless the create_xml() method is called, followed at some point with model.store_epc()
        """

        if np.all(self.block_dxyz_dkji[0, :2] == 0.0):
            return 'vertical'
        return 'straight'

    def xyz_box(grid, points_root = None, lazy = True, local = False):
        """Returns the minimum and maximum xyz for the grid geometry.

        arguments:
           points_root (ignored): for compatibility with Grid method signature
           lazy (ignored): for compatibility with Grid method signature
           local (boolean, default False): if True, the xyz ranges that are returned are in the local
              coordinate space, otherwise the global (crs parent) coordinate space

        returns:
           numpy array of float of shape (2, 3); the first axis is minimum, maximum; the second axis is x, y, z

        :meta common:
        """

        if grid.xyz_box_cached is None:
            grid.xyz_box_cached = np.empty((2, 3))
            if grid.is_aligned:
                dxyz = np.array([grid.block_dxyz_dkji[2 - axis, axis] for axis in range(3)])
                temp_box = np.zeros((2, 3))
                temp_box[1] = np.array((grid.ni, grid.nj, grid.nk), dtype = float) * dxyz
                grid.xyz_box_cached[0] = np.amin(temp_box, axis = 0)
                grid.xyz_box_cached[1] = np.amax(temp_box, axis = 0)
            else:
                # generate points for outer grid corners, find min & max by axis
                grid_cp = np.zeros((2, 2, 2, 3))
                for k in [0, 1]:
                    lcp_k = float(k * grid.nk) * grid.block_dxyz_dkji[0]
                    for j in [0, 1]:
                        lcp_j = lcp_k + float(j * grid.nj) * grid.block_dxyz_dkji[1]
                        for i in [0, 1]:
                            grid_cp[k, j, i] = lcp_j + float(i * grid.ni) * grid.block_dxyz_dkji[2]
                grid_cp = grid_cp.reshape((8, 3))
                grid.xyz_box_cached[0] = np.amin(grid_cp, axis = 0)
                grid.xyz_box_cached[1] = np.amax(grid_cp, axis = 0)
            grid.xyz_box_cached_thoroughly = True
        if local:
            return grid.xyz_box_cached
        global_xyz_box = grid.xyz_box_cached.copy()
        grid.local_to_global_crs(global_xyz_box, crs_uuid = grid.crs_uuid)
        g_xyz_box = np.empty((2, 3), dtype = float)
        g_xyz_box[0] = np.amin(global_xyz_box, axis = 0)
        g_xyz_box[1] = np.amax(global_xyz_box, axis = 0)
        return g_xyz_box

    def find_cell_for_point_xy(self, x, y, k0 = 0, vertical_ref = 'top', local_coords = True):
        """Searches in 2D for a cell containing point x,y in layer k0; return (j0, i0) or (None, None)."""

        if x is None or y is None:
            return (None, None)
        if not local_coords and self.crs_uuid is not None:
            a = np.array([[x, y, 0.0]])  # extra axis needed to keep global_to_local_crs happy
            self.global_to_local_crs(a, crs_uuid = self.crs_uuid)
            x = a[0, 0]
            y = a[0, 1]
        log.debug(f'find for local x: {x}; y: {y}')
        if self.is_aligned:
            i0 = maths.floor(x / self.block_dxyz_dkji[2, 0])
            j0 = maths.floor(y / self.block_dxyz_dkji[1, 1])
        else:
            # TODO
            raise NotImplementedError('finding cell for x,y in unaligned regular grid')
            i0 = j0 = -1
        log.debug(f'found j0: {j0}; i0: {i0}')
        if 0 <= i0 < self.ni and 0 <= j0 < self.nj:
            return (j0, i0)
        return (None, None)

    def create_xml(self,
                   ext_uuid = None,
                   add_as_part = True,
                   add_relationships = True,
                   set_as_grid_root = True,
                   title = None,
                   originator = None,
                   write_active = True,
                   write_geometry = None,
                   extra_metadata = {},
                   expand_const_arrays = False,
                   add_cell_length_properties = True,
                   use_lattice = True):
        """Creates xml for this RegularGrid object; by default the explicit geometry is not included.

        see docstring for Grid.create_xml()

        additional argument:
           add_cell_length_properties (boolean, default True): if True, 3 constant property arrays with cells as
              indexable element are created to hold the lengths of the primary axes of the cells; the xml is
              created for the properties and they are added to the model (no hdf5 write needed)
           use_lattice (boolean, default True): if True, a geometry node is created with a Point3dLatticeArray;

        note:
           if the regular grid has been instatiated using the as_irregular_grid mode, a lattice will not be used,
           regardless of the use_lattice argument;
           if using a lattice representation, the write_geomety argument is ignored and a lattice geometry will
           be created

        :meta common:
        """

        if extra_metadata is None:
            extra_metadata = {}

        if write_geometry is None:
            write_geometry = (self.grid_representation == 'IjkGrid')

        if self.grid_representation == 'IjkGrid':
            use_lattice = False

        if write_geometry is None:
            write_geometry = (self.grid_representation == 'IjkGrid')
        else:
            extra_metadata = extra_metadata.copy()

        if write_geometry and not use_lattice:
            extra_metadata['grid_flavour'] = 'IjkGrid'
        if write_geometry or use_lattice:
            if 'crs uuid' in extra_metadata:
                extra_metadata.pop('crs uuid')
        elif self.crs_uuid is not None:
            extra_metadata['crs uuid'] = str(self.crs_uuid)

        node = super().create_xml(ext_uuid = ext_uuid,
                                  add_as_part = add_as_part,
                                  add_relationships = add_relationships,
                                  set_as_grid_root = set_as_grid_root,
                                  title = title,
                                  originator = originator,
                                  write_active = write_active,
                                  write_geometry = write_geometry or use_lattice,
                                  extra_metadata = extra_metadata,
                                  use_lattice = use_lattice)
        assert node is not None

        # patch in constant ...is_defined arrays if required
        if write_geometry or use_lattice:
            geom = rqet.find_tag(node, 'Geometry')
            assert geom is not None
            if always_write_pillar_geometry_is_defined and rqet.find_tag(geom, 'PillarGeometryIsDefined') is None:
                grr_xml._add_constant_pillar_geometry_is_defined(geom, self.extent_kji)
            if always_write_cell_geometry_is_defined and rqet.find_tag(geom, 'CellGeometryIsDefined') is None:
                grr_xml._add_constant_cell_geometry_is_defined(geom, self.extent_kji)

        if add_cell_length_properties:
            axes_lengths_kji = self.axial_lengths_kji()
            dpc = rprop.GridPropertyCollection()
            dpc.set_grid(self)
            for axis in range(3):
                dpc.add_cached_array_to_imported_list(None,
                                                      'regular grid',
                                                      'D' + 'ZYX'[axis],
                                                      discrete = False,
                                                      uom = self.xy_units(),
                                                      property_kind = 'cell length',
                                                      facet_type = 'direction',
                                                      facet = 'KJI'[axis],
                                                      indexable_element = 'cells',
                                                      count = 1,
                                                      const_value = axes_lengths_kji[axis])
            if expand_const_arrays:
                dpc.write_hdf5_for_imported_list(expand_const_arrays = True)
            dpc.create_xml_for_imported_list_and_add_parts_to_model(expand_const_arrays = expand_const_arrays)
            if self.property_collection is None:
                self.property_collection = dpc
            else:
                if self.property_collection.support is None:
                    self.property_collection.set_support(support = self)
                self.property_collection.inherit_parts_from_other_collection(dpc)

            if add_relationships and self.crs_uuid is not None:
                self.model.create_reciprocal_relationship_uuids(self.uuid, 'destinationObject', self.crs_uuid,
                                                                'sourceObject')

        return node

    def _set_is_aligned(self):
        """Sets is_aligned attribute True if IJK axes align with +/- xyz respectively."""
        if self.block_dxyz_dkji is None:
            self.is_aligned = None
        self.is_aligned = (np.count_nonzero(self.block_dxyz_dkji * np.array([[1.0, 1, 0], [1, 0, 1], [0, 1, 1]])) == 0)

    def aligned_dxyz(self):
        """Returns triple float dx, dy, dz cell dimensions for aligned grids only."""
        assert self.is_aligned
        return (self.block_dxyz_dkji[2, 0], self.block_dxyz_dkji[1, 1], self.block_dxyz_dkji[0, 2])

    def slice_points(self, axis = 0, ref_slice = 0, local = False):
        """Returns a slice of points data for a layer or a cross section.

        arguments:
           axis (int, default 0): the axis of slicing 0 for K, 1 for J, 2 for I
           ref_slice (int, default 0): the slice index in the direction of axis
           local (bool, default False): if True, returned points are in the local crs; if False in global

        returns:
           numpy float array of points data of shape (Na, Nb, 3); (Na, Nb) are (NJ+1, NI+1), (NK+1, NI+1) or
           (NK+1, NJ+1) for axis values of 0, 1, or 2 respectively
        """
        assert 0 <= axis < 3
        assert 0 <= ref_slice <= self.extent_kji[axis]
        other_axes = np.array([[1, 2], [0, 2], [0, 1]], dtype = int)[axis]
        shape = (self.extent_kji[other_axes[0]] + 1, self.extent_kji[other_axes[1]] + 1, 3)
        dxyz_da = self.block_dxyz_dkji[other_axes[0]]
        dxyz_db = self.block_dxyz_dkji[other_axes[1]]
        p = np.zeros(shape, dtype = float)
        if ref_slice:
            p[0, 0] = ref_slice * self.block_dxyz_dkji[axis]
        for a in range(1, shape[0]):
            p[a, 0] = p[0, 0] + a * dxyz_da
        for b in range(1, shape[1]):
            p[:, b] = p[:, 0] + b * dxyz_db
        if not local:
            assert self.crs is not None
            self.crs.local_to_global_array(p)
        return p

    def _add_geom_points_xml(self, geom_node, ext_uuid):
        # RegularGrid override method using lattice representation, called from Grid.create_xml()

        def make_axial_offset(model, p_node, axis, d, n):
            dxyz = np.zeros(3, dtype = float)
            dxyz[2 - axis] = d
            o_node = rqet.SubElement(p_node, ns['resqml2'] + 'Offset')
            o_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dOffset')
            o_node.text = '\n'
            model.create_solitary_point3d('Offset', o_node, dxyz)
            space_node = rqet.SubElement(o_node, ns['resqml2'] + 'Spacing')
            space_node.set(ns['xsi'] + 'type',
                           ns['resqml2'] + 'DoubleConstantArray')  # nothing else catered for just now
            space_node.text = '\n'
            ov_node = rqet.SubElement(space_node, ns['resqml2'] + 'Value')
            ov_node.set(ns['xsi'] + 'type', ns['xsd'] + 'double')
            ov_node.text = '1'
            oc_node = rqet.SubElement(space_node, ns['resqml2'] + 'Count')
            oc_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
            oc_node.text = str(n)

        assert self.crs_uuid is not None

        if self.grid_representation == 'IjkBlockGrid':

            # todo: handle non aligned regular grids
            assert self.is_aligned

            points_node = rqet.SubElement(geom_node, ns['resqml2'] + 'Points')
            points_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dLatticeArray')
            points_node.text = '\n'

            adao = rqet.SubElement(points_node, ns['resqml2'] + 'AllDimensionsAreOrthogonal')
            adao.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
            adao.text = 'true'

            self.model.create_solitary_point3d('Origin', points_node, (0.0, 0.0, 0.0))

            for axis in range(3):  # note: 1st Offset is for K axis, 2nd for J, then I
                make_axial_offset(self.model, points_node, axis, self.aligned_dxyz()[2 - axis], self.extent_kji[axis])

        else:  # generate geometry as irregular grid

            super()._add_geom_points_xml(geom_node, ext_uuid)



"""GridSkin class for representing outer skin of a Grid."""

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.surface as rqs
import resqpy.grid_surface as rqgs


class GridSkin:
    """Class of object consisting of outer skin of grid (not a RESQML class in its own right)."""

    def __init__(self, grid, quad_triangles = True, use_single_layer_tactics = True, is_regular = False):
        """Returns a composite surface object consisting of outer skin of grid."""

        if grid.k_gaps:
            use_single_layer_tactics = False

        self.grid = grid
        self.k_gaps = 0 if grid.k_gaps is None else grid.k_gaps
        self.k_gap_after_layer_list = [
        ]  # list of layer numbers (zero based) where there is a gap after (ie. below, usually)
        self.has_split_coordinate_lines = grid.has_split_coordinate_lines
        self.quad_triangles = quad_triangles
        self.use_single_layer_tactics = use_single_layer_tactics
        self.skin = None  #: compostite surface constructed during initialisation
        self.fault_j_face_cols_ji0 = None  # split internal J faces
        self.fault_i_face_cols_ji0 = None  # split internal I faces
        self.polygon = None  # not yet in use
        self.is_regular = is_regular  #: indicates a simplified skin for a regular aligned grid

        k_gap_surf_list = self._make_k_gap_surfaces(quad_triangles = quad_triangles)

        if self.is_regular:

            # build a simplified two triangle surface for each of the six skin surfaces
            xyz_box = grid.xyz_box(local = True)
            for axis in range(3):
                if grid.block_dxyz_dkji[2 - axis, axis] < 0.0:
                    xyz_box[:, axis] = (xyz_box[1, axis], xyz_box[0, axis])
            min_x, min_y, min_z = xyz_box[0]
            max_x, max_y, max_z = xyz_box[1]
            top_surf = rqs.Surface(grid.model)
            top_surf.set_to_horizontal_plane(0.0, xyz_box)
            base_surf = rqs.Surface(grid.model)
            base_surf.set_to_horizontal_plane(grid.nk * grid.block_dxyz_dkji[0, 2], xyz_box)
            j_minus_surf = rqs.Surface(grid.model)
            corners = np.array([(min_x, min_y, min_z), (max_x, min_y, min_z), (min_x, min_y, max_z),
                                (max_x, min_y, max_z)])
            j_minus_surf.set_to_triangle_pair(corners)
            j_plus_surf = rqs.Surface(grid.model)
            corners = np.array([(min_x, max_y, min_z), (max_x, max_y, min_z), (min_x, max_y, max_z),
                                (max_x, max_y, max_z)])
            j_plus_surf.set_to_triangle_pair(corners)
            i_minus_surf = rqs.Surface(grid.model)
            corners = np.array([(min_x, min_y, min_z), (min_x, max_y, min_z), (min_x, min_y, max_z),
                                (min_x, max_y, max_z)])
            i_minus_surf.set_to_triangle_pair(corners)
            i_plus_surf = rqs.Surface(grid.model)
            corners = np.array([(max_x, min_y, min_z), (max_x, max_y, min_z), (max_x, min_y, max_z),
                                (max_x, max_y, max_z)])
            i_plus_surf.set_to_triangle_pair(corners)
            surf_list = [top_surf, base_surf, j_minus_surf, j_plus_surf, i_minus_surf, i_plus_surf]

        elif self.has_split_coordinate_lines:

            top_surf = rqgs.generate_torn_surface_for_layer_interface(grid,
                                                                      k0 = 0,
                                                                      ref_k_faces = 'top',
                                                                      quad_triangles = quad_triangles)
            base_surf = rqgs.generate_torn_surface_for_layer_interface(grid,
                                                                       k0 = grid.nk - 1,
                                                                       ref_k_faces = 'base',
                                                                       quad_triangles = quad_triangles)
            j_minus_surf = rqgs.generate_torn_surface_for_x_section(grid,
                                                                    'J',
                                                                    ref_slice0 = 0,
                                                                    plus_face = False,
                                                                    quad_triangles = quad_triangles,
                                                                    as_single_layer = use_single_layer_tactics)
            j_plus_surf = rqgs.generate_torn_surface_for_x_section(grid,
                                                                   'J',
                                                                   ref_slice0 = self.grid.nj - 1,
                                                                   plus_face = True,
                                                                   quad_triangles = quad_triangles,
                                                                   as_single_layer = use_single_layer_tactics)
            i_minus_surf = rqgs.generate_torn_surface_for_x_section(grid,
                                                                    'I',
                                                                    ref_slice0 = 0,
                                                                    plus_face = False,
                                                                    quad_triangles = quad_triangles,
                                                                    as_single_layer = use_single_layer_tactics)
            i_plus_surf = rqgs.generate_torn_surface_for_x_section(grid,
                                                                   'I',
                                                                   ref_slice0 = self.grid.ni - 1,
                                                                   plus_face = True,
                                                                   quad_triangles = quad_triangles,
                                                                   as_single_layer = use_single_layer_tactics)

            # fault face processing
            j_column_faces, i_column_faces = grid.split_column_faces()
            col_j0, col_i0 = np.where(j_column_faces)
            self.fault_j_face_cols_ji0 = np.stack((col_j0, col_i0),
                                                  axis = -1)  # fault is on plus j face of these columns
            col_j0, col_i0 = np.where(i_column_faces)
            self.fault_i_face_cols_ji0 = np.stack((col_j0, col_i0),
                                                  axis = -1)  # fault is on plus i face of these columns

            j_minus_fault_surf = j_plus_fault_surf = None
            j_minus_fault_surf_list = []
            j_plus_fault_surf_list = []
            for col_ji0 in self.fault_j_face_cols_ji0:
                #            log.debug(f'fault j face col_ji0: {col_ji0}')
                # note: use of single layer for internal fault surfaces results in some incorrect skin surface where k gaps are present
                # find_first_intersection_of_trajectory() method takes care of this but other uses of skin might need to be aware
                _, surf = rqgs.create_column_face_mesh_and_surface(grid,
                                                                   col_ji0,
                                                                   1,
                                                                   1,
                                                                   quad_triangles = quad_triangles,
                                                                   as_single_layer = True)
                j_plus_fault_surf_list.append(surf)
                _, surf = rqgs.create_column_face_mesh_and_surface(grid, (col_ji0[0] + 1, col_ji0[1]),
                                                                   1,
                                                                   0,
                                                                   quad_triangles = quad_triangles,
                                                                   as_single_layer = True)
                j_minus_fault_surf_list.append(surf)
            if len(j_minus_fault_surf_list) > 0:
                j_minus_fault_surf = rqs.CombinedSurface(j_minus_fault_surf_list, grid.crs_uuid)
            if len(j_plus_fault_surf_list) > 0:
                j_plus_fault_surf = rqs.CombinedSurface(j_plus_fault_surf_list, grid.crs_uuid)

            i_minus_fault_surf = i_plus_fault_surf = None
            i_minus_fault_surf_list = []
            i_plus_fault_surf_list = []
            for col_ji0 in self.fault_i_face_cols_ji0:
                #            log.debug(f'fault i face col_ji0: {col_ji0}')
                _, surf = rqgs.create_column_face_mesh_and_surface(grid,
                                                                   col_ji0,
                                                                   2,
                                                                   1,
                                                                   quad_triangles = quad_triangles,
                                                                   as_single_layer = True)
                i_plus_fault_surf_list.append(surf)
                _, surf = rqgs.create_column_face_mesh_and_surface(grid, (col_ji0[0], col_ji0[1] + 1),
                                                                   2,
                                                                   0,
                                                                   quad_triangles = quad_triangles,
                                                                   as_single_layer = True)
                i_minus_fault_surf_list.append(surf)
            if len(i_minus_fault_surf_list) > 0:
                i_minus_fault_surf = rqs.CombinedSurface(i_minus_fault_surf_list, grid.crs_uuid)
            if len(i_plus_fault_surf_list) > 0:
                i_plus_fault_surf = rqs.CombinedSurface(i_plus_fault_surf_list, grid.crs_uuid)

            surf_list = [top_surf, base_surf, j_minus_surf, j_plus_surf, i_minus_surf, i_plus_surf]
            for k_gap_surf in k_gap_surf_list:
                surf_list.append(k_gap_surf)
            for fault_surf in [j_minus_fault_surf, j_plus_fault_surf, i_minus_fault_surf, i_plus_fault_surf]:
                if fault_surf is not None:
                    surf_list.append(fault_surf)

        else:

            top_surf = rqgs.generate_untorn_surface_for_layer_interface(grid,
                                                                        k0 = 0,
                                                                        ref_k_faces = 'top',
                                                                        quad_triangles = quad_triangles)
            base_surf = rqgs.generate_untorn_surface_for_layer_interface(grid,
                                                                         k0 = grid.nk - 1,
                                                                         ref_k_faces = 'base',
                                                                         quad_triangles = quad_triangles)
            j_minus_surf = rqgs.generate_untorn_surface_for_x_section(grid,
                                                                      'J',
                                                                      ref_slice0 = 0,
                                                                      plus_face = False,
                                                                      quad_triangles = quad_triangles,
                                                                      as_single_layer = use_single_layer_tactics)
            j_plus_surf = rqgs.generate_untorn_surface_for_x_section(grid,
                                                                     'J',
                                                                     ref_slice0 = self.grid.nj - 1,
                                                                     plus_face = True,
                                                                     quad_triangles = quad_triangles,
                                                                     as_single_layer = use_single_layer_tactics)
            i_minus_surf = rqgs.generate_untorn_surface_for_x_section(grid,
                                                                      'I',
                                                                      ref_slice0 = 0,
                                                                      plus_face = False,
                                                                      quad_triangles = quad_triangles,
                                                                      as_single_layer = use_single_layer_tactics)
            i_plus_surf = rqgs.generate_untorn_surface_for_x_section(grid,
                                                                     'I',
                                                                     ref_slice0 = self.grid.ni - 1,
                                                                     plus_face = True,
                                                                     quad_triangles = quad_triangles,
                                                                     as_single_layer = use_single_layer_tactics)

            surf_list = [top_surf, base_surf, j_minus_surf, j_plus_surf, i_minus_surf, i_plus_surf]
            for k_gap_surf in k_gap_surf_list:
                surf_list.append(k_gap_surf)

        self.skin = rqs.CombinedSurface(surf_list)

    def find_first_intersection_of_trajectory(self,
                                              trajectory,
                                              start = 0,
                                              start_xyz = None,
                                              nudge = None,
                                              exclude_kji0 = None):
        """Returns the first intersection of the trajectory with the torn skin.

        Returns the x,y,z and K,J,I and axis, polarity & segment.

        arguments:
           trajectory (well.Trajectory object): the trajectory to be intersected with the skin
           start (int, default 0): the trajectory segment number to start the search from
           start_xyz (triple float, optional): if present, this point should lie on the start segment and search continues from
              this point
           nudge (float, optional): if present and positive, the start point is nudged forward by this distance (grid uom);
              if present and negative (more typical for skin entry search), the start point is nudged back a little
           exclude_kji0 (triple int, optional): if present, the indices of a cell to exclude as a possible result

        returns:
           5-tuple of:

           - (triple float): xyz coordinates of the intersection point in the crs of the grid
           - (triple int): kji0 of the cell that is intersected, which might be a pinched out or otherwise inactive cell
           - (int): 0, 1 or 2 for K, J or I axis of cell face
           - (int): 0 for -ve face, 1 for +ve face
           - (int): trajectory knot prior to the intersection (also the segment number)

        note:
           if the GridSkin object has been initialised using single layer tactics, then the k0 value will be zero for any
           initial entry through a sidewall of the grid or through a fault face
        """

        if exclude_kji0 is not None:
            bundle = rqgs.find_first_intersection_of_trajectory_with_surface(trajectory,
                                                                             self.skin,
                                                                             start = start,
                                                                             start_xyz = start_xyz,
                                                                             nudge = nudge,
                                                                             return_second = True)
            xyz_1, segment_1, tri_index_1, xyz_2, segment_2, tri_index_2 = bundle
        else:
            xyz_1, segment_1, tri_index_1 = rqgs.find_first_intersection_of_trajectory_with_surface(
                trajectory, self.skin, start = start, start_xyz = start_xyz, nudge = nudge)
            xyz_2 = segment_2 = tri_index_2 = None
        if xyz_1 is None:
            return None, None, None, None, None
        if xyz_2 is None:
            triplet_list = [(xyz_1, segment_1, tri_index_1)]
        else:
            triplet_list = [(xyz_1, segment_1, tri_index_1), (xyz_2, segment_2, tri_index_2)]

        for (try_xyz, segment, tri_index) in triplet_list:

            xyz = try_xyz

            surf_index, surf_tri_index = self.skin.surface_index_for_triangle_index(tri_index)
            assert surf_index is not None

            if self.is_regular:
                assert 0 <= surf_index < 6
                axis, polarity = divmod(surf_index, 2)
                kji0 = np.zeros(3, dtype = int)
                if polarity:
                    kji0[axis] = self.grid.extent_kji[axis] - 1
                if axis == 0:  # K face (top or base)
                    kji0[1] = int(xyz[1] / self.grid.block_dxyz_dkji[1, 1])
                    kji0[2] = int(xyz[0] / self.grid.block_dxyz_dkji[2, 0])
                elif axis == 1:  # J face
                    kji0[0] = int(xyz[2] / self.grid.block_dxyz_dkji[0, 2])
                    kji0[2] = int(xyz[0] / self.grid.block_dxyz_dkji[2, 0])
                else:  # I face
                    kji0[0] = int(xyz[2] / self.grid.block_dxyz_dkji[0, 2])
                    kji0[1] = int(xyz[1] / self.grid.block_dxyz_dkji[1, 1])
                kji0 = tuple(kji0)

            elif surf_index < 6:  # grid skin
                # following returns j,i pair for K faces; k,i for J faces; or k,j for I faces
                col = self.skin.surface_list[surf_index].column_from_triangle_index(surf_tri_index)
                if surf_index == 0:
                    kji0 = (0, col[0], col[1])  # K- (top)
                elif surf_index == 1:
                    kji0 = (self.grid.nk - 1, col[0], col[1])  # K+ (base)
                elif surf_index == 2:
                    kji0 = (col[0], 0, col[1])  # J-
                elif surf_index == 3:
                    kji0 = (col[0], self.grid.nj - 1, col[1])  # J+
                elif surf_index == 4:
                    kji0 = (col[0], col[1], 0)  # I-
                else:
                    kji0 = (col[0], col[1], self.grid.ni - 1)  # I+
                axis, polarity = divmod(surf_index, 2)
                if self.use_single_layer_tactics and surf_index > 1:  # now compare against layered representation of column face
                    xyz, k0 = rqgs.find_intersection_of_trajectory_interval_with_column_face(trajectory,
                                                                                             self.grid,
                                                                                             segment,
                                                                                             kji0[1:],
                                                                                             axis,
                                                                                             polarity,
                                                                                             start_xyz = xyz,
                                                                                             nudge = -1.0)
                    if xyz is None:
                        log.error('unexpected failure to identify skin column face penetration point')
                        return None, None, None, None, None
                    kji0 = (k0, kji0[1], kji0[2])

            elif surf_index < 6 + 2 * self.k_gaps:  # top or base face of a k gap
                col = self.skin.surface_list[surf_index].column_from_triangle_index(surf_tri_index)
                surf_index -= 6
                gap_index, top_base = divmod(surf_index, 2)
                axis = 0
                if top_base == 0:  # top of k gap (ie. base of layer before)
                    kji0 = (self.k_gap_after_layer_list[gap_index], col[0], col[1])
                    polarity = 1
                else:  # base of k gap (top of layer after)
                    kji0 = (self.k_gap_after_layer_list[gap_index] + 1, col[0], col[1])
                    polarity = 0

            else:  # fault face
                axis, polarity = divmod(surf_index - (6 + 2 * self.k_gaps), 2)
                axis += 1
                log.debug('penetration through fault face ' + 'KJI'[axis] + '-+'[polarity])
                assert self.skin.is_combined_list[surf_index]
                fault_surf_list = self.skin.surface_list[surf_index]
                col_surf_index, _ = fault_surf_list.surface_index_for_triangle_index(surf_tri_index)
                col_ji0 = np.empty((2,), dtype = int)
                surf_index -= 6 + 2 * self.k_gaps
                if surf_index in [0, 1]:  # J-/+ fault face
                    col_ji0[:] = self.fault_j_face_cols_ji0[col_surf_index]
                elif surf_index in [2, 3]:  # I-/+ fault face
                    col_ji0[:] = self.fault_i_face_cols_ji0[col_surf_index]
                else:  # should not be possible
                    raise Exception('code failure')
                if polarity == 0:
                    col_ji0[axis - 1] += 1
                xyz_f, k0 = rqgs.find_intersection_of_trajectory_interval_with_column_face(trajectory,
                                                                                           self.grid,
                                                                                           segment,
                                                                                           col_ji0,
                                                                                           axis,
                                                                                           polarity,
                                                                                           start_xyz = xyz,
                                                                                           nudge = -1.0)
                if xyz_f is None:
                    if self.k_gaps:
                        log.debug('failed to identify fault penetration point; assumed to be in k gap; nudging forward')
                        if nudge is None or nudge < 0.0:
                            nudge = 0.0
                        nudge += 0.1
                        return self.find_first_intersection_of_trajectory(trajectory,
                                                                          start = segment,
                                                                          start_xyz = xyz,
                                                                          nudge = nudge,
                                                                          exclude_kji0 = exclude_kji0)
                    log.error('unexpected failure to identify fault penetration point')
                    return None, None, None, None, None
                xyz = xyz_f
                kji0 = (k0, col_ji0[0], col_ji0[1])

            if exclude_kji0 is None or not np.all(kji0 == exclude_kji0):
                return xyz, kji0, axis, polarity, segment

        return None, None, None, None, None

    def _make_k_gap_surfaces(self, quad_triangles = True):
        """Returns a list of newly created surfaces representing top and base of each k gap."""

        # list of layer face surfaces, 2 per gap (top of gap, ie. base of layer above, then base of gap)
        k_gap_surf_list = []
        self.k_gap_after_layer_list = []
        if self.k_gaps > 0:
            for k0 in range(self.grid.nk - 1):
                if self.grid.k_gap_after_array[k0]:
                    if self.has_split_coordinate_lines:
                        gap_top_surf = rqgs.generate_torn_surface_for_layer_interface(self.grid,
                                                                                      k0 = k0,
                                                                                      ref_k_faces = 'base',
                                                                                      quad_triangles = quad_triangles)
                        gap_base_surf = rqgs.generate_torn_surface_for_layer_interface(self.grid,
                                                                                       k0 = k0 + 1,
                                                                                       ref_k_faces = 'top',
                                                                                       quad_triangles = quad_triangles)
                    else:
                        gap_top_surf = rqgs.generate_untorn_surface_for_layer_interface(self.grid,
                                                                                        k0 = k0,
                                                                                        ref_k_faces = 'base',
                                                                                        quad_triangles = quad_triangles)
                        gap_base_surf = rqgs.generate_untorn_surface_for_layer_interface(
                            self.grid, k0 = k0 + 1, ref_k_faces = 'top', quad_triangles = quad_triangles)
                    k_gap_surf_list.append(gap_top_surf)
                    k_gap_surf_list.append(gap_base_surf)
                    self.k_gap_after_layer_list.append(k0)
        assert len(k_gap_surf_list) == 2 * self.k_gaps
        assert len(self.k_gap_after_layer_list) == self.k_gaps
        return k_gap_surf_list



"""Copy grid function."""

import numpy as np

import resqpy.crs as rqc
import resqpy.grid as grr


def copy_grid(source_grid, target_model = None, copy_crs = True):
    """Creates a copy of the IJK grid object in the target model (usually prior to modifying points in situ).

    note:
       this function is not usually called directly by application code; it does not write to the hdf5
       file nor create xml for the copied grid;
       the copy will be a resqpy Grid even if the source grid is a RegularGrid
    """

    grid_flavour = grr.grid_flavour(source_grid.root)
    assert grid_flavour in ['IjkGrid', 'IjkBlockGrid']

    model = source_grid.model
    if target_model is None:
        target_model = model
    if target_model is model:
        copy_crs = False

    # if the source grid is a RegularGrid, ensure that it has explicit points
    if grid_flavour == 'IjkBlockGrid':
        source_grid.make_regular_points_cached()

    # create empty grid object (with new uuid)
    grid = grr.Grid(target_model)

    # inherit attributes from source grid (but a RegularGrid is copied to a Grid)
    grid.grid_representation = 'IjkGrid'
    grid.extent_kji = np.array(source_grid.extent_kji, dtype = 'int')
    grid.nk, grid.nj, grid.ni = source_grid.nk, source_grid.nj, source_grid.ni
    grid.k_direction_is_down = source_grid.k_direction_is_down
    grid.grid_is_right_handed = source_grid.grid_is_right_handed
    grid.pillar_shape = source_grid.pillar_shape
    grid.has_split_coordinate_lines = source_grid.has_split_coordinate_lines
    grid.k_gaps = source_grid.k_gaps
    if grid.k_gaps:
        grid.k_gap_after_array = source_grid.k_gap_after_array.copy()
        grid.k_raw_index_array = source_grid.k_raw_index_array.copy()

    # inherit a copy of the coordinate reference system used by the grid geometry
    grid.crs_uuid = source_grid.crs_uuid
    if target_model is source_grid.model:
        grid.crs = rqc.Crs(model, uuid = grid.crs_uuid)
    elif copy_crs and source_grid.crs_uuid is not None:
        model.duplicate_node(source_grid.model.root_for_uuid(source_grid.crs_uuid), add_as_part = True)
    else:
        grid.crs = None

    # inherit a copy of the inactive cell mask
    if source_grid.inactive is None:
        grid.inactive = None
    else:
        grid.inactive = source_grid.inactive.copy()
    grid.active_property_uuid = source_grid.active_property_uuid

    # take a copy of the grid geometry
    source_grid.cache_all_geometry_arrays()
    grid.geometry_defined_for_all_pillars_cached = source_grid.geometry_defined_for_all_pillars_cached
    if hasattr(source_grid,
               'array_pillar_geometry_is_defined') and source_grid.array_pillar_geometry_is_defined is not None:
        grid.array_pillar_geometry_is_defined = source_grid.array_pillar_geometry_is_defined.copy()
    if hasattr(source_grid,
               'array_cell_geometry_is_defined') and source_grid.array_cell_geometry_is_defined is not None:
        grid.array_cell_geometry_is_defined = source_grid.array_cell_geometry_is_defined.copy()
    grid.geometry_defined_for_all_cells_cached = source_grid.geometry_defined_for_all_cells_cached
    grid.points_cached = source_grid.points_cached.copy()
    if grid.has_split_coordinate_lines:
        source_grid.create_column_pillar_mapping()
        grid.split_pillar_indices_cached = source_grid.split_pillar_indices_cached.copy()
        grid.cols_for_split_pillars = source_grid.cols_for_split_pillars.copy()
        grid.cols_for_split_pillars_cl = source_grid.cols_for_split_pillars_cl.copy()
        grid.split_pillars_count = source_grid.split_pillars_count
        grid.pillars_for_column = source_grid.pillars_for_column.copy()

    return grid



"""_import_vdb_ensemble.py: Module to import a vdb ensemble into resqml format."""

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import resqpy.model as rq
import resqpy.olio.vdb as vdb
import resqpy.olio.xml_et as rqet
import resqpy.property as rp
import resqpy.rq_import as rqi
import resqpy.time_series as rts


def import_vdb_ensemble(
        epc_file,
        ensemble_run_dir,
        existing_epc = False,
        keyword_list = None,
        property_kind_list = None,
        vdb_static_properties = True,  # if True, static vdb properties are imported
        vdb_recurrent_properties = True,
        decoarsen = True,
        timestep_selection = 'all',
        create_property_set_per_realization = True,
        create_property_set_per_timestep = True,
        create_complete_property_set = False,
        # remaining arguments only used if existing_epc is False
        extent_ijk = None,  # 3 element numpy vector
        corp_xy_units = 'm',
        corp_z_units = 'm',
        corp_z_inc_down = True,
        ijk_handedness = 'right',
        geometry_defined_everywhere = True,
        treat_as_nan = None,
        resqml_xy_units = 'm',
        resqml_z_units = 'm',
        resqml_z_inc_down = True,
        shift_to_local = True,
        local_origin_place = 'centre',  # 'centre' or 'minimum'
        max_z_void = 0.1,  # import will fail if vertical void greater than this is encountered
        split_pillars = True,
        split_tolerance = 0.01,  # applies to each of x, y, z differences
        progress_fn = None):
    """Adds properties from all vdb's within an ensemble directory tree to a single RESQML dataset.

    Referencing a shared grid.

    args:
       epc_file (string): filename of epc file to be extended with ensemble properties
       ensemble_run_dir (string): path of main ensemble run directory; vdb's within this directory tree are source of import
       existing_epc (boolean, default False): if True, the epc_file must already exist and contain the compatible grid
       keyword_list (list of strings, optional): if present, only properties for keywords within the list are included
       property_kind_list (list of strings, optional): if present, only properties which are mapped to these resqml property
          kinds are included in the import
       vdb_static_properties (boolean, default True): if False, no static properties are included, regardless of keyword and/or
          property kind matches
       vdb_recurrent_properties (boolean, default True): if False, no recurrent properties are included, regardless of keyword
          and/or property kind matches
       decoarsen (boolean, default True): if True and ICOARSE property exists for a grid in a case, the associated property
          data is decoarsened; if False, the property data is as stored in the vdb
       timestep_selection (string, default 'all'): may be 'first', 'last', 'first and last', or 'all', controlling which
          reporting timesteps are included when loading recurrent data
       create_property_set_per_realization (boolean, default True): if True, a property set object is created for each realization
       create_property_set_per_timestep (boolean, default True): if True, a property set object is created for each timestep
          included in the recurrent data import
       create_complete_property_set (boolean, default False): if True, a property set object is created containing all the
          properties imported; only really useful to differentiate from other properties related to the grid
       extent_ijk (triple int, optional): this and remaining arguments are only used if existing_epc is False; the extent
          is only needed in case automatic determination of the extent fails
       corp_xy_units (string, default 'm'): the units of x & y values in the vdb corp data;
          typically 'm' (metres), 'ft' (feet) or 'cm' (centimetres, for lab scale models)
       corp_z_units (string, default 'm'): the units of z values in the vdb corp data;
          typically 'm' (metres), 'ft' (feet) or 'cm' (centimetres, for lab scale models)
       corp_z_inc_down (boolean, default True): set to True if corp z values are depth; False if elevation
       ijk_handedness (string, default 'right'): set to the handedness of the IJK axes in the Nexus model; 'right' or 'left'
       geometry_defined_everywhere (boolean, default True): set to False if inactive cells do not have valid geometry;
          deprecated - use treat_as_nan argument instead
       treat_as_nan (string, optional): if not None, one of 'dots', 'ij_dots', 'inactive'; controls which inactive cells
          have their geometry set to undefined
       resqml_xy_units (string, default 'm'): the units of x & y values to use in the generated resqml grid;
          typically 'm' (metres), 'ft' (feet) or 'cm' (centimetres, for lab scale models)
       resqml_z_units (string, default 'm'): the units of z values to use in the generated resqml grid;
          typically 'm' (metres), 'ft' (feet) or 'cm' (centimetres, for lab scale models)
       resqml_z_inc_down (boolean, default True): set to True if resqml z values are to be depth; False for elevations
       shift_to_local (boolean, default True): if True, the resqml coordinate reference system will use a local origin
       local_origin_place (string, default 'centre'): where to place the local origin; 'centre' or 'minimum'; only
          relevant if shift_to_local is True
       max_z_void (float, default 0.1): the tolerance of voids between layers, in z direction; voids greater than this
          will cause the grid import to fail
       split_pillars (boolean, default True): if False, a grid is generated without split pillars
       split_tolerance (float, default 0.01): the tolerance applied to each of x, y, & z values, beyond which a corner
          point (and hence pillar) will be split
       progress_fn (function(float), optional): if present, this function is called at intervals during processing; it
          must accept one floating point argument which will range from 0.0 to 1.0

    returns:
       resqpy.Model object containing properties for all the realisations; hdf5 and epc files having been updated

    note:
       if existing_epc is True, the epc file must already exist and contain one grid (or one grid named ROOT) which must
       have the correct extent for all realisations within the ensemble; if existing_epc is False, the resqml dataset is
       created afresh with a grid extracted from the first realisation in the ensemble; either way, the single grid is used
       as the representative grid in the ensemble resqml dataset being generated;
       all vdb directories within the directory tree headed by ensemble_run_dir are included in the import; by
       default all properties will be imported; the keyword_list, property_kind_list, vdb_static_properties,
       vdb_recurrent_properties and timestep_selection arguments can be used to filter the required properties;
       if both keyword_list and property_kind_list are provided, a property must match an item in both lists in order
       to be included; if recurrent properties are being included then all vdb's should contain the same number of reporting
       steps in their recurrent data and these should relate to the same set of timestamps; timestamp data is extracted from a
       summary file for the first realisation; no check is made to ensure that reporting timesteps in different realisations
       are actually for the same date.
    """

    assert epc_file.endswith('.epc')
    assert vdb_static_properties or vdb_recurrent_properties, 'no properties selected for ensemble import'

    if progress_fn is not None:
        progress_fn(0.0)

    # fetch a sorted list of the vdb paths found in the run directory tree
    ensemble_list = vdb.ensemble_vdb_list(ensemble_run_dir)
    if len(ensemble_list) == 0:
        log.error("no vdb's found in run directory tree: " + str(ensemble_run_dir))
        return None

    if not existing_epc:
        model = rqi.import_nexus(
            epc_file[:-4],  # output path and file name without .epc or .h5 extension
            extent_ijk = extent_ijk,  # 3 element numpy vector, in case extent is not automatically determined
            vdb_file = ensemble_list[0],  # vdb input file
            corp_xy_units = corp_xy_units,
            corp_z_units = corp_z_units,
            corp_z_inc_down = corp_z_inc_down,
            ijk_handedness = ijk_handedness,
            geometry_defined_everywhere = geometry_defined_everywhere,
            treat_as_nan = treat_as_nan,
            resqml_xy_units = resqml_xy_units,
            resqml_z_units = resqml_z_units,
            resqml_z_inc_down = resqml_z_inc_down,
            shift_to_local = shift_to_local,
            local_origin_place = local_origin_place,  # 'centre' or 'minimum'
            max_z_void = max_z_void,  # import will fail if vertical void greater than this is encountered
            split_pillars = split_pillars,
            split_tolerance = split_tolerance,  # applies to each of x, y, z differences
            vdb_static_properties = False,
            vdb_recurrent_properties = False,
            create_property_set = False)

    model = rq.Model(
        epc_file = epc_file)  # shouldn't be necessary if just created but it feels safer to re-open the model
    assert model is not None, 'failed to instantiate model'
    grid = model.grid()
    assert grid is not None, 'grid not found'
    ext_uuid = model.h5_uuid()
    assert ext_uuid is not None, 'failed to determine uuid for hdf5 file reference'
    hdf5_file = model.h5_file_name(uuid = ext_uuid)

    # create reporting timestep time series for recurrent data, if required, based on the first realisation
    recur_time_series = None
    recur_ts_uuid = None
    timestep_list = None
    if vdb_recurrent_properties:
        summary_file = ensemble_list[0][:-4] + '.sum'  # TODO: check timestep summary file extension, .tssum?
        full_time_series = rts.time_series_from_nexus_summary(summary_file)
        if full_time_series is None:
            log.error('failed to extract info from timestep summary file; disabling recurrent property import')
            vdb_recurrent_properties = False
    if vdb_recurrent_properties:
        vdbase = vdb.VDB(ensemble_list[0])
        timestep_list = vdbase.list_of_timesteps()
        if len(timestep_list) == 0:
            log.warning(
                'no ROOT recurrent data found in vdb for first realisation; disabling recurrent property import')
            vdb_recurrent_properties = False
    if vdb_recurrent_properties:
        if timestep_selection == 'all' or ('first' in timestep_selection):
            fs_index = 0
        else:
            fs_index = -1
        first_stamp = full_time_series.timestamp(timestep_list[fs_index])
        if first_stamp is None:
            log.error('first timestamp number selected for import was not found in summary file: ' +
                      str(timestep_list[fs_index]))
            log.error('disabling recurrent property import')
            vdb_recurrent_properties = False
    if vdb_recurrent_properties:
        recur_time_series = rts.TimeSeries(model, first_timestamp = first_stamp)
        if timestep_selection == 'all':
            remaining_list = timestep_list[1:]
        elif timestep_selection == 'first and last':
            remaining_list = [timestep_list[-1]]
        else:
            remaining_list = []
        for timestep_number in remaining_list:
            stamp = full_time_series.timestamp(timestep_number)
            if stamp is None:
                log.error('timestamp number for which recurrent data exists was not found in summary file: ' +
                          str(timestep_number))
                log.error('disabling recurrent property import')
                vdb_recurrent_properties = False
                recur_time_series = None
                break
            recur_time_series.add_timestamp(stamp)
    if recur_time_series is not None:
        recur_ts_node = recur_time_series.create_xml(title = 'simulator recurrent array timestep series')
        recur_ts_uuid = rqet.uuid_for_part_root(recur_ts_node)
        model.time_series = recur_ts_node  # save as the primary time series for the model

    if create_complete_property_set or create_property_set_per_timestep:
        complete_collection = rp.GridPropertyCollection()
        complete_collection.set_grid(grid)
    else:
        complete_collection = None

    #  main loop over realisations

    for realisation in range(len(ensemble_list)):

        if progress_fn is not None:
            progress_fn(float(1 + realisation) / float(1 + len(ensemble_list)))

        vdb_file = ensemble_list[realisation]
        log.info('processing realisation ' + str(realisation) + ' from: ' + str(vdb_file))
        vdbase = vdb.VDB(vdb_file)
        #      case_list = vdbase.cases()
        #      assert len(case_list) > 0, 'no cases found in vdb: ' + str(vdb_file)
        #      if len(case_list) > 1: log.warning('more than one case found in vdb (using first): ' + str(vdb_file))
        #      vdb_case = case_list[0]
        #      vdbase.set_use_case(vdb_case)
        vdbase.set_extent_kji(grid.extent_kji)

        prop_import_collection = rp.GridPropertyCollection(realization = realisation)
        prop_import_collection.set_grid(grid)

        decoarsen_array = None
        if vdb_static_properties:
            props = vdbase.list_of_static_properties()
            if len(props) > 0:
                for keyword in props:
                    if keyword_list is not None and keyword not in keyword_list:
                        continue
                    prop_kind, facet_type, facet = rp.property_kind_and_facet_from_keyword(keyword)
                    if property_kind_list is not None and prop_kind not in property_kind_list and prop_kind not in [
                            'active', 'region initialization'
                    ]:
                        continue
                    prop_import_collection.import_vdb_static_property_to_cache(vdbase,
                                                                               keyword,
                                                                               realization = realisation,
                                                                               property_kind = prop_kind,
                                                                               facet_type = facet_type,
                                                                               facet = facet)
                if decoarsen:
                    decoarsen_array = prop_import_collection.decoarsen_imported_list()
                    if decoarsen_array is not None:
                        log.debug('static properties decoarsened for realisation ' + str(realisation))
                grid.write_hdf5_from_caches(hdf5_file,
                                            mode = 'a',
                                            geometry = False,
                                            imported_properties = prop_import_collection,
                                            write_active = False)
                prop_import_collection.remove_all_cached_arrays()

        if vdb_recurrent_properties:

            r_timestep_list = vdbase.list_of_timesteps()  # get list of timesteps for which recurrent files exist
            if len(r_timestep_list) < recur_time_series.number_of_timestamps():
                log.error('insufficient number of reporting timesteps; skipping recurrent data for realisation ' +
                          str(realisation))
            else:
                common_recur_prop_set = None
                for tni in range(recur_time_series.number_of_timestamps()):
                    if timestep_selection in ['all', 'first']:
                        timestep_number = timestep_list[tni]
                        r_timestep_number = r_timestep_list[tni]
                    elif timestep_selection == 'last' or tni > 0:
                        timestep_number = timestep_list[-1]
                        r_timestep_number = r_timestep_list[-1]
                    else:
                        timestep_number = timestep_list[0]
                        r_timestep_number = r_timestep_list[0]
                    stamp = full_time_series.timestamp(timestep_number)
                    recur_prop_list = vdbase.list_of_recurrent_properties(r_timestep_number)
                    if common_recur_prop_set is None:
                        common_recur_prop_set = set(recur_prop_list)
                    elif recur_prop_list is not None:
                        common_recur_prop_set = common_recur_prop_set.intersection(set(recur_prop_list))
                    step_import_collection = rp.GridPropertyCollection()
                    step_import_collection.set_grid(grid)
                    # for each property for this timestep, cache array and add to recur prop import collection for this time step
                    if recur_prop_list:
                        for keyword in recur_prop_list:
                            if not keyword or not keyword.isalnum():
                                continue
                            if keyword_list is not None and keyword not in keyword_list:
                                continue
                            prop_kind, facet_type, facet = rp.property_kind_and_facet_from_keyword(keyword)
                            if property_kind_list is not None and prop_kind not in property_kind_list:
                                continue
                            step_import_collection.import_vdb_recurrent_property_to_cache(
                                vdbase,
                                r_timestep_number,
                                keyword,
                                time_index = tni,  # index into recur_time_series
                                realization = realisation,
                                property_kind = prop_kind,
                                facet_type = facet_type,
                                facet = facet)
                    if decoarsen_array is not None:
                        step_import_collection.decoarsen_imported_list(decoarsen_array = decoarsen_array)
                    # extend hdf5 with cached arrays for this timestep
                    #         log.info('number of recurrent grid property arrays for timestep: ' + str(timestep_number) +
                    #                  ' is: ' + str(step_import_collection.number_of_imports()))
                    #         log.info('extending hdf5 file with recurrent properties for timestep: ' + str(timestep_number))
                    grid.write_hdf5_from_caches(hdf5_file,
                                                mode = 'a',
                                                geometry = False,
                                                imported_properties = step_import_collection,
                                                write_active = False)
                    # add imported list for this timestep to full imported list
                    prop_import_collection.inherit_imported_list_from_other_collection(step_import_collection)
                    #         log.debug('total number of property arrays after timestep: ' + str(timestep_number) +
                    #                   ' is: ' + str(prop_import_collection.number_of_imports()))
                    # remove cached copies of arrays
                    step_import_collection.remove_all_cached_arrays()

        if len(prop_import_collection.imported_list) == 0:
            log.warning('no properties imported for realisation ' + str(realisation))
            continue

        prop_import_collection.create_xml_for_imported_list_and_add_parts_to_model(ext_uuid,
                                                                                   time_series_uuid = recur_ts_uuid)

        if create_property_set_per_realization:
            prop_import_collection.create_property_set_xml('property set for realization ' + str(realisation))

        if complete_collection is not None:
            complete_collection.inherit_parts_from_other_collection(prop_import_collection)

    if complete_collection is not None:
        if create_property_set_per_timestep and recur_time_series is not None:
            for tni in range(recur_time_series.number_of_timestamps()):
                ts_collection = rp.selective_version_of_collection(complete_collection, time_index = tni)
                if ts_collection.number_of_parts() > 0:
                    ts_collection.create_property_set_xml('property set for time index ' + str(tni))
        if create_complete_property_set:
            complete_collection.create_property_set_xml('property set for ensemble vdb import')

    # mark model as modified (will already have happened anyway)
    model.set_modified()

    # rewrite epc file
    log.info('storing updated model in epc file ' + epc_file)
    model.store_epc(epc_file)

    if progress_fn is not None:
        progress_fn(1.0)

    # return updated resqml model
    return model



"""Unstructured grid and derived classes."""

__all__ = [
    'UnstructuredGrid', 'HexaGrid', 'PrismGrid', 'VerticalPrismGrid', 'PyramidGrid', 'TetraGrid', 'valid_cell_shapes'
]

from ._unstructured_grid import UnstructuredGrid, valid_cell_shapes
from ._hexa_grid import HexaGrid
from ._prism_grid import PrismGrid, VerticalPrismGrid
from ._pyramid_grid import PyramidGrid
from ._tetra_grid import TetraGrid

# Set "module" attribute of all public objects to this path. Fixes issue #310
for _name in __all__:
    _obj = eval(_name)
    if hasattr(_obj, "__module__"):
        _obj.__module__ = __name__



"""Multiprocessing wrapper functions for the grid_surface module."""

import logging

log = logging.getLogger(__name__)

import numpy as np
import uuid
from typing import Tuple, Union, List, Optional, Callable
from pathlib import Path
from uuid import UUID

import resqpy.grid_surface as rqgs
import resqpy.model as rq
import resqpy.grid as grr
import resqpy.property as rqp
import resqpy.surface as rqs
import resqpy.olio.uuid as bu


def find_faces_to_represent_surface_regular_wrapper(
        index: int,
        parent_tmp_dir: str,
        use_index_as_realisation: bool,
        grid_epc: str,
        grid_uuid: Union[UUID, str],
        surface_epc: str,
        surface_uuid: Union[UUID, str],
        name: str,
        title: Optional[str] = None,
        agitate: bool = False,
        random_agitation: bool = False,
        feature_type: str = 'fault',
        trimmed: bool = False,
        is_curtain = False,
        extend_fault_representation: bool = False,
        flange_inner_ring = False,
        saucer_parameter = None,
        retriangulate: bool = False,
        related_uuid = None,
        progress_fn: Optional[Callable] = None,
        extra_metadata = None,
        return_properties: Optional[List[str]] = None,
        raw_bisector: bool = False,
        use_pack: bool = False) -> Tuple[int, bool, str, List[Union[UUID, str]]]:
    """Multiprocessing wrapper function of find_faces_to_represent_surface_regular_optimised.

    arguments:
        index (int): the index of the function call from the multiprocessing function
        parent_tmp_dir (str): the parent temporary directory path from the multiprocessing function
        use_index_as_realisation (bool): if True, uses the index number as the realization number on
            the property collection
        grid_epc (str): epc file path where the grid is saved
        grid_uuid (UUID or str): UUID (universally unique identifier) of the grid object
        surface_epc (str): epc file path where the surface (or point set) is saved
        surface_uuid (UUID or str): UUID (universally unique identifier) of the surface (or point set) object.
        name (str): the feature name to use in the grid connection set.
        title (str): the citation title to use for the grid connection set; defaults to name
        agitate (bool): if True, the points of the surface are perturbed by a small offset,
           which can help if the surface has been built from a regular mesh with a periodic resonance
           with the grid
        random_agitation (bool, default False): if True, the agitation is by a small random distance; if False,
           a constant positive shift of 5.0e-6 is applied to x, y & z values; ignored if agitate is False
        feature_type (str, default 'fault'): one of 'fault', 'horizon', or 'geobody boundary'
        trimmed (bool, default True): if True the surface has already been trimmed
        is_curtain (bool, default False): if True, only the top layer is intersected with the surface and bisector
           is generated as a column property if requested
        extend_fault_representation (bool, default False): if True, the representation is extended with a flange
        retriangulate (bool, default False): if True, a retriangulation is performed even if not needed otherwise
        related_uuid (uuid, optional): if present, the uuid of an object to be softly related to the gcs (and to
           grid bisector and/or shadow property if requested)
        progress_fn (Callable): a callback function to be called at intervals by this function;
           the argument will progress from 0.0 to 1.0 in unspecified and uneven increments
        extra_metadata (dict, optional): extra metadata items to be added to the grid connection set
        return_properties (list of str): if present, a list of property arrays to calculate and
           return as a dictionary; recognised values in the list are 'triangle', 'depth', 'offset', 'normal vector',
           'flange bool', 'grid bisector' and 'grid shadow';
           triangle is an index into the surface triangles of the triangle detected for the gcs face; depth is
           the z value of the intersection point of the inter-cell centre vector with a triangle in the surface;
           offset is a measure of the distance between the centre of the cell face and the intersection point;
           normal vector is a unit vector normal to the surface triangle; each array has an entry for each face
           in the gcs; grid bisector is a grid cell boolean property holding True for the set of cells on one
           side of the surface, deemed to be shallower; grid shadow is a grid cell int8 property holding 1 for
           cells neither above nor below a K face, 1 for above, 2 for beneath, 3 for between;
           the returned dictionary has the passed strings as keys and numpy arrays as values
        raw_bisector (bool, default False): if True and grid bisector is requested then it is left in a raw
           form without assessing which side is shallower (True values indicate same side as origin cell)
        use_pack (bool, default False): if True, boolean properties will be stored in numpy packed format,
           which will only be readable by resqpy based applications

    returns:
        Tuple containing:
            - index (int): the index passed to the function
            - success (bool): whether the function call was successful, whatever that definiton is
            - epc_file (str): the epc file path where the objects are stored
            - uuid_list (List[str]): list of UUIDs of relevant objects

    notes:
        Use this function as argument to the multiprocessing function; it will create a new model that is saved
        in a temporary epc file and returns the required values, which are used in the multiprocessing function to
        recombine all the objects into a single epc file
    """
    tmp_dir = Path(parent_tmp_dir) / f"{uuid.uuid4()}"
    tmp_dir.mkdir(parents = True, exist_ok = True)
    epc_file = str(tmp_dir / "wrapper.epc")
    model = rq.new_model(epc_file = epc_file, quiet = True)
    uuid_list = []
    g_model = rq.Model(grid_epc, quiet = True)
    g_crs_uuid = g_model.uuid(obj_type = 'LocalDepth3dCrs',
                              related_uuid = grid_uuid)  # todo: check this relationship exists
    if g_crs_uuid is not None:
        model.copy_uuid_from_other_model(g_model, g_crs_uuid)
        uuid_list.append(g_crs_uuid)
    model.copy_uuid_from_other_model(g_model, uuid = grid_uuid)
    uuid_list.append(grid_uuid)
    for prop_title in ['DX', 'DY', 'DZ']:
        prop_uuid = g_model.uuid(obj_type = 'ContinuousProperty', title = prop_title, related_uuid = grid_uuid)
        if prop_uuid is not None:
            log.debug(f'found grid cell length property {prop_title}')
            model.copy_uuid_from_other_model(g_model, uuid = prop_uuid)
            uuid_list.append(prop_uuid)
        else:
            log.warning(f'grid cell length property {prop_title} NOT found')
    grid = grr.RegularGrid(parent_model = model, uuid = grid_uuid)
    assert grid.is_aligned
    flange_radius = 5.0 * np.sum(np.array(grid.extent_kji, dtype = float) * np.array(grid.aligned_dxyz()))
    s_model = rq.Model(surface_epc, quiet = True)
    model.copy_uuid_from_other_model(s_model, uuid = str(surface_uuid))
    repr_type = model.type_of_part(model.part(uuid = surface_uuid), strip_obj = True)
    assert repr_type in ['TriangulatedSetRepresentation', 'PointSetRepresentation']
    extended = False
    retriangulated = False
    flange_bool = None
    if repr_type == 'PointSetRepresentation':
        # trim pointset to grid xyz box
        pset = rqs.PointSet(model, uuid = surface_uuid)
        surf_title = pset.title
        log.debug(f'point set {pset.title} raw point count: {len(pset.full_array_ref())}')
        pset.change_crs(grid.crs)
        if not trimmed:
            pset.trim_to_xyz_box(grid.xyz_box(local = True))
            trimmed = True
            if 'trimmed' not in surf_title:
                surf_title += ' trimmed'
        assert len(pset.full_array_ref()) >= 3,  \
            f'boundary {name} representation {pset.title} has no xyz overlap with grid'
        pset_points = pset.full_array_ref()
        log.debug(f'trimmed point set {pset.title} contains {len(pset_points)} points; about to triangulate')
        if len(pset_points) > 50_000:
            log.warning(
                f'trimmed point set {pset.title} has {len(pset_points)} points, which might take a while to triangulate'
            )
        # triangulate point set to form a surface; set repr_uuid to that surface and switch repr_flavour to 'surface'
        if extend_fault_representation and not surf_title.endswith(' extended'):
            surf_title += ' extended'
        surf = rqs.Surface(model, crs_uuid = grid.crs.uuid, title = surf_title)
        flange_bool = surf.set_from_point_set(pset,
                                              convexity_parameter = 2.0,
                                              reorient = True,
                                              extend_with_flange = extend_fault_representation,
                                              flange_inner_ring = flange_inner_ring,
                                              saucer_parameter = saucer_parameter,
                                              flange_radial_distance = flange_radius,
                                              make_clockwise = False)
        extended = extend_fault_representation
        retriangulated = True
        surf.write_hdf5()
        surf.create_xml()
        inherit_interpretation_relationship(model, surface_uuid, surf.uuid)
        surface_uuid = surf.uuid

    surface = rqs.Surface(parent_model = model, uuid = str(surface_uuid))
    surf_title = surface.title
    assert surf_title
    surface.change_crs(grid.crs)
    if not trimmed and surface.triangle_count() > 100:
        if not surf_title.endswith('trimmed'):
            surf_title += ' trimmed'
        trimmed_surf = rqs.Surface(model, crs_uuid = grid.crs.uuid, title = surf_title)
        # trimmed_surf.set_to_trimmed_surface(surf, xyz_box = xyz_box, xy_polygon = parent_seg.polygon)
        trimmed_surf.set_to_trimmed_surface(surface, xyz_box = grid.xyz_box(local = True))
        surface = trimmed_surf
        trimmed = True
    if (extend_fault_representation and not extended) or (retriangulate and not retriangulated):
        _, p = surface.triangles_and_points()
        pset = rqs.PointSet(model, points_array = p, crs_uuid = grid.crs.uuid, title = surf_title)
        if extend_fault_representation and not surf_title.endswith('extended'):
            surf_title += ' extended'
        surface = rqs.Surface(model, crs_uuid = grid.crs.uuid, title = surf_title)
        flange_bool = surface.set_from_point_set(pset,
                                                 convexity_parameter = 2.0,
                                                 reorient = True,
                                                 extend_with_flange = extend_fault_representation,
                                                 flange_inner_ring = flange_inner_ring,
                                                 saucer_parameter = saucer_parameter,
                                                 flange_radial_distance = flange_radius,
                                                 make_clockwise = False)
        del pset
        extended = extend_fault_representation
        retriangulated = True
    if not bu.matching_uuids(surface.uuid, surface_uuid):
        surface.write_hdf5()
        surface.create_xml()
        # relate modified surface to original
        model.create_reciprocal_relationship_uuids(surface.uuid, 'sourceObject', surface_uuid, 'destinationObject')
        # inherit relationship to an interpretation object, if present for original surface
        inherit_interpretation_relationship(model, surface_uuid, surface.uuid)
        surface_uuid = surface.uuid
    if flange_bool is not None:
        flange_p = rqp.Property.from_array(parent_model = model,
                                           cached_array = flange_bool,
                                           source_info = 'flange bool array',
                                           keyword = 'flange bool',
                                           support_uuid = surface_uuid,
                                           property_kind = 'flange bool',
                                           find_local_property_kind = True,
                                           indexable_element = 'faces',
                                           discrete = True)
        uuid_list.append(flange_p.uuid)
    uuid_list.append(surface_uuid)

    returns = rqgs.find_faces_to_represent_surface_regular_optimised(grid,
                                                                     surface,
                                                                     name,
                                                                     title,
                                                                     agitate,
                                                                     random_agitation,
                                                                     feature_type,
                                                                     is_curtain,
                                                                     progress_fn,
                                                                     return_properties,
                                                                     raw_bisector = raw_bisector)

    success = False

    if isinstance(returns, tuple):
        gcs = returns[0]
    else:
        gcs = returns

    if gcs.count > 0:
        success = True
        gcs.write_hdf5()
        gcs.create_xml(extra_metadata = extra_metadata)
        model.copy_uuid_from_other_model(gcs.model, uuid = gcs.uuid)
        if related_uuid is not None:
            relative_found = (model.uuid(uuid = related_uuid) is not None)
            if not relative_found:
                for m in gcs.model, g_model, s_model:
                    if m.uuid(uuid = related_uuid) is not None:
                        model.copy_uuid_from_other_model(m, related_uuid)
                        relative_found = True
                        break
            if relative_found:
                model.create_reciprocal_relationship_uuids(gcs.uuid, 'sourceObject', related_uuid, 'destinationObject')
            else:
                log.warning(f'related uuid {related_uuid} not found; relationship dropped')
        uuid_list.append(gcs.uuid)

    if success and return_properties is not None and len(return_properties):
        log.debug(f'{name} requested properties: {return_properties}')
        properties = returns[1]
        realisation = index if use_index_as_realisation else None
        property_collection = rqp.PropertyCollection(support = gcs)
        grid_pc = None
        for p_name, array in properties.items():
            log.debug(f'{name} found property {p_name}')
            if p_name == "normal vector":
                property_collection.add_cached_array_to_imported_list(
                    array,
                    f"from find_faces function for {surface.title}",
                    f'{surface.title} {p_name}',
                    discrete = False,
                    uom = "Euc",
                    property_kind = "normal vector",
                    realization = realisation,
                    indexable_element = "faces",
                    points = True,
                )
            elif p_name == "triangle":
                property_collection.add_cached_array_to_imported_list(
                    array,
                    f"from find_faces function for {surface.title}",
                    f'{surface.title} {p_name}',
                    discrete = True,
                    null_value = -1,
                    property_kind = "triangle index",
                    realization = realisation,
                    indexable_element = "faces",
                )
            elif p_name == "offset":
                property_collection.add_cached_array_to_imported_list(
                    array,
                    f"from find_faces function for {surface.title}",
                    f'{surface.title} {p_name}',
                    discrete = False,
                    uom = grid.crs.z_units,
                    property_kind = "offset",
                    realization = realisation,
                    indexable_element = "faces",
                )
            elif p_name == "depth":
                # convert values to global z inc down
                array[:] += grid.crs.z_offset
                if not grid.crs.z_inc_down:
                    array = -array
                property_collection.add_cached_array_to_imported_list(
                    array,
                    f"from find_faces function for {surface.title}",
                    f'{surface.title} {p_name}',
                    discrete = False,
                    uom = grid.crs.z_units,
                    property_kind = "depth",
                    realization = realisation,
                    indexable_element = "faces",
                )
            elif p_name == 'grid bisector':
                array, is_curtain = array
                if grid_pc is None:
                    grid_pc = rqp.PropertyCollection()
                    grid_pc.set_support(support = grid)
                grid_pc.add_cached_array_to_imported_list(
                    array,
                    f"from find_faces function for {surface.title}",
                    f'{surface.title} {p_name}',
                    discrete = True,
                    property_kind = "grid bisector",
                    facet_type = 'direction',
                    facet = 'raw' if raw_bisector else ('vertical' if is_curtain else 'sloping'),
                    realization = realisation,
                    indexable_element = "columns" if is_curtain else "cells",
                )
            elif p_name == 'grid shadow':
                if grid_pc is None:
                    grid_pc = rqp.PropertyCollection()
                    grid_pc.set_support(support = grid)
                grid_pc.add_cached_array_to_imported_list(
                    array,
                    f"from find_faces function for {surface.title}",
                    f'{surface.title} {p_name}',
                    discrete = True,
                    property_kind = "grid shadow",
                    realization = realisation,
                    indexable_element = "cells",
                )
            elif p_name == 'flange bool':
                property_collection.add_cached_array_to_imported_list(
                    array,
                    f"from find_faces function for {surface.title}",
                    f'{surface.title} {p_name}',
                    discrete = True,
                    null_value = -1,
                    property_kind = "flange bool",
                    realization = realisation,
                    indexable_element = "faces",
                )
            else:
                raise ValueError(f'unrecognised property name {p_name}')
        if property_collection.number_of_imports() > 0:
            # log.debug('writing gcs property hdf5 data')
            property_collection.write_hdf5_for_imported_list(use_pack = use_pack)
            uuids_properties = property_collection.create_xml_for_imported_list_and_add_parts_to_model(
                find_local_property_kinds = True)
            uuid_list.extend(uuids_properties)
        if grid_pc is not None and grid_pc.number_of_imports() > 0:
            # log.debug('writing grid property (bisector) hdf5 data')
            grid_pc.write_hdf5_for_imported_list(use_pack = use_pack)
            # log.debug('creating xml for grid property (bisector)')
            uuids_properties = grid_pc.create_xml_for_imported_list_and_add_parts_to_model(
                find_local_property_kinds = True)
            assert uuids_properties
            uuid_list.extend(uuids_properties)
            if related_uuid is not None:
                for p_uuid in uuids_properties:
                    # log.debug(f'creating relationship between: {p_uuid} and {related_uuid}')
                    model.create_reciprocal_relationship_uuids(p_uuid, 'sourceObject', related_uuid,
                                                               'destinationObject')
    else:
        log.debug(f'{name} no requested properties')

    # log.debug('find_faces_to_represent_surface_regular_wrapper() storing epc: {model.epc_file}')
    model.store_epc(quiet = True)

    return index, success, epc_file, uuid_list


def inherit_interpretation_relationship(model, old_repr_uuid, new_repr_uuid):
    """Inherit a relationship to an interpretation object if present for old representation."""
    for obj_type in [
            'HorizonInterpretation', 'FaultInterpretation', 'BoundaryFeatureInterpretation',
            'GeobodyBoundaryInterpretation'
    ]:
        interp_uuid = model.uuid(obj_type = obj_type, related_uuid = old_repr_uuid)
        if interp_uuid is not None:
            model.create_reciprocal_relationship_uuids(new_repr_uuid, 'sourceObject', interp_uuid, 'destinationObject')
            break



"""RESQML manipulation library.

.. autosummary::
    :toctree: _autosummary
    :caption: API Reference
    :template: custom-module-template.rst
    :recursive:

    model
    crs
    derived_model
    fault
    grid
    grid_surface
    lines
    multi_processing
    organize
    property
    rq_import
    strata
    surface
    time_series
    unstructured
    weights_and_measures
    well
    olio
"""

import logging

__version__ = "4.11.1"  # Set at build time
log = logging.getLogger(__name__)
log.info(f"Imported resqpy version {__version__}")
